"""
***************************************************************************
    ParametersPanel.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
                           (C) 2013 by CS Systemes d'information (CS SI)
    Email                : volayaf at gmail dot com
                           otb at c-s dot fr (CS SI)
    Contributors         : Victor Olaya

***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

from qgis.core import (
    Qgis,
    QgsProcessingModelAlgorithm,
    QgsProcessingOutputLayerDefinition,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterExtent,
    QgsProject,
)
from qgis.gui import (
    QgsAbstractProcessingParameterWidgetWrapper,
    QgsGui,
    QgsProcessingContextGenerator,
    QgsProcessingGui,
    QgsProcessingHiddenWidgetWrapper,
    QgsProcessingParametersGenerator,
    QgsProcessingParametersWidget,
    QgsProcessingParameterWidgetContext,
)
from qgis.PyQt.QtWidgets import QLabel, QSizePolicy, QVBoxLayout, QWidget
from qgis.utils import iface

from processing.core.exceptions import InvalidOutputExtension, InvalidParameterValue
from processing.gui.wrappers import WidgetWrapper, WidgetWrapperFactory
from processing.tools.dataobjects import createContext


class ParametersPanel(QgsProcessingParametersWidget):
    def __init__(self, parent, alg, in_place=False, active_layer=None):
        super().__init__(alg, parent)
        self.in_place = in_place

        self.active_layer = active_layer

        self.wrappers = {}

        self.extra_parameters = {}

        self.processing_context = createContext()

        class ContextGenerator(QgsProcessingContextGenerator):
            def __init__(self, context):
                super().__init__()
                self.processing_context = context

            def processingContext(self):
                return self.processing_context

        self.context_generator = ContextGenerator(self.processing_context)

        self.initWidgets()

        QgsProject.instance().layerWasAdded.connect(self.layerRegistryChanged)
        QgsProject.instance().layersWillBeRemoved.connect(self.layerRegistryChanged)

    def layerRegistryChanged(self, layers):
        for wrapper in list(self.wrappers.values()):
            try:
                wrapper.refresh()
            except AttributeError:
                pass

    def initWidgets(self):
        super().initWidgets()

        widget_context = QgsProcessingParameterWidgetContext()
        widget_context.setProject(QgsProject.instance())
        if iface is not None:
            widget_context.setMapCanvas(iface.mapCanvas())
            widget_context.setBrowserModel(iface.browserModel())
            widget_context.setActiveLayer(iface.activeLayer())

        widget_context.setMessageBar(self.parent().messageBar())
        if isinstance(self.algorithm(), QgsProcessingModelAlgorithm):
            widget_context.setModel(self.algorithm())

        in_place_input_parameter_name = "INPUT"
        if hasattr(self.algorithm(), "inputParameterName"):
            in_place_input_parameter_name = self.algorithm().inputParameterName()

        # If there are no parameters to show because it's in-place, we add the info label.
        # Still, it needs the following steps to create the parameter widgets (even when hidden).
        if self.in_place and not [
            d
            for d in self.algorithm().parameterDefinitions()
            if d.name() not in (in_place_input_parameter_name, "OUTPUT")
        ]:
            widget = QWidget(self)
            layout = QVBoxLayout(widget)
            label = QLabel(widget)
            label.setWordWrap(True)
            info_text = self.tr(
                "<i>No additional parameters are required. This algorithm will activate edit mode and modify the features on layer <b>{layername}</b> in place.</i>"
            ).format(layername=self.active_layer.name())
            label.setText(info_text)
            layout.addWidget(label)
            layout.addStretch()
            self.addExtraWidget(widget)

        # Create widgets and put them in layouts
        for param in self.algorithm().parameterDefinitions():
            if param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                continue

            if param.isDestination():
                continue
            else:
                if self.in_place and param.name() in (
                    in_place_input_parameter_name,
                    "OUTPUT",
                ):
                    # don't show the input/output parameter widgets in in-place mode
                    # we still need to CREATE them, because other wrappers may need to interact
                    # with them (e.g. those parameters which need the input layer for field
                    # selections/crs properties/etc)
                    self.wrappers[param.name()] = QgsProcessingHiddenWidgetWrapper(
                        param, QgsProcessingGui.WidgetType.Standard, self
                    )
                    self.wrappers[param.name()].setLinkedVectorLayer(self.active_layer)
                    continue

                wrapper = WidgetWrapperFactory.create_wrapper(param, self.parent())
                wrapper.setWidgetContext(widget_context)
                wrapper.registerProcessingContextGenerator(self.context_generator)
                wrapper.registerProcessingParametersGenerator(self)
                self.wrappers[param.name()] = wrapper

                # For compatibility with 3.x API, we need to check whether the wrapper is
                # the deprecated WidgetWrapper class. If not, it's the newer
                # QgsAbstractProcessingParameterWidgetWrapper class
                # TODO QGIS 5.0 - remove
                is_python_wrapper = issubclass(wrapper.__class__, WidgetWrapper)
                stretch = 0
                if not is_python_wrapper:
                    widget = wrapper.createWrappedWidget(self.processing_context)
                    wrapper.widgetValueHasChanged.connect(self.parameterChanged)
                    stretch = wrapper.stretch()
                else:
                    widget = wrapper.widget

                if widget is not None:
                    if is_python_wrapper:
                        widget.setToolTip(param.toolTip())

                    label = None
                    if not is_python_wrapper:
                        label = wrapper.createWrappedLabel()
                    else:
                        label = wrapper.label

                    if label is not None:
                        self.addParameterLabel(param, label)
                    elif is_python_wrapper:
                        desc = param.description()
                        if isinstance(param, QgsProcessingParameterExtent):
                            desc += self.tr(" (xmin, xmax, ymin, ymax)")
                        if (
                            param.flags()
                            & QgsProcessingParameterDefinition.Flag.FlagOptional
                        ):
                            desc += self.tr(" [optional]")
                        widget.setText(desc)

                    self.addParameterWidget(param, widget, stretch)

        for output in self.algorithm().destinationParameterDefinitions():
            if output.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                continue

            if self.in_place and output.name() in (
                in_place_input_parameter_name,
                "OUTPUT",
            ):
                continue

            wrapper = QgsGui.processingGuiRegistry().createParameterWidgetWrapper(
                output, QgsProcessingGui.WidgetType.Standard
            )
            wrapper.setWidgetContext(widget_context)
            wrapper.registerProcessingContextGenerator(self.context_generator)
            wrapper.registerProcessingParametersGenerator(self)
            self.wrappers[output.name()] = wrapper

            label = wrapper.createWrappedLabel()
            if label is not None:
                self.addOutputLabel(label)

            widget = wrapper.createWrappedWidget(self.processing_context)
            self.addOutputWidget(widget, wrapper.stretch())

            #    def skipOutputChanged(widget, checkbox, skipped):
            # TODO
            #        enabled = not skipped
            #
            #        # Do not try to open formats that are write-only.
            #        value = widget.value()
            #        if value and isinstance(value, QgsProcessingOutputLayerDefinition) and isinstance(output, (
            #                QgsProcessingParameterFeatureSink, QgsProcessingParameterVectorDestination)):
            #            filename = value.sink.staticValue()
            #            if filename not in ('memory:', ''):
            #                path, ext = os.path.splitext(filename)
            #                format = QgsVectorFileWriter.driverForExtension(ext)
            #                drv = gdal.GetDriverByName(format)
            #                if drv:
            #                    if drv.GetMetadataItem(gdal.DCAP_OPEN) is None:
            #                        enabled = False
            #
            #        checkbox.setEnabled(enabled)
            #        checkbox.setChecked(enabled)

        for wrapper in list(self.wrappers.values()):
            wrapper.postInitialize(list(self.wrappers.values()))

    def createProcessingParameters(
        self, flags=QgsProcessingParametersGenerator.Flags()
    ):
        include_default = not (
            flags & QgsProcessingParametersGenerator.Flag.SkipDefaultValueParameters
        )
        validate = not (flags & QgsProcessingParametersGenerator.Flag.SkipValidation)
        parameters = {}
        for p, v in self.extra_parameters.items():
            parameters[p] = v

        for param in self.algorithm().parameterDefinitions():
            if param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                continue
            if not param.isDestination():
                try:
                    wrapper = self.wrappers[param.name()]
                except KeyError:
                    continue

                # For compatibility with 3.x API, we need to check whether the wrapper is
                # the deprecated WidgetWrapper class. If not, it's the newer
                # QgsAbstractProcessingParameterWidgetWrapper class
                # TODO QGIS 5.0 - remove
                if issubclass(wrapper.__class__, WidgetWrapper):
                    widget = wrapper.widget
                else:
                    widget = wrapper.wrappedWidget()

                if (
                    not isinstance(wrapper, QgsProcessingHiddenWidgetWrapper)
                    and widget is None
                ):
                    continue

                value = wrapper.parameterValue()
                if param.defaultValue() != value or include_default:
                    parameters[param.name()] = value

                if validate and not param.checkValueIsAcceptable(value):
                    raise InvalidParameterValue(param, widget)
            else:
                if self.in_place and param.name() == "OUTPUT":
                    parameters[param.name()] = "memory:"
                    continue

                try:
                    wrapper = self.wrappers[param.name()]
                except KeyError:
                    continue

                widget = wrapper.wrappedWidget()
                value = wrapper.parameterValue()

                dest_project = None
                if wrapper.customProperties().get("OPEN_AFTER_RUNNING"):
                    dest_project = QgsProject.instance()

                if value and isinstance(value, QgsProcessingOutputLayerDefinition):
                    value.destinationProject = dest_project
                if value and (param.defaultValue() != value or include_default):
                    parameters[param.name()] = value

                    context = createContext()
                    if validate:
                        ok, error = param.isSupportedOutputValue(value, context)
                        if not ok:
                            raise InvalidOutputExtension(widget, error)

        return self.algorithm().preprocessParameters(parameters)

    def setParameters(self, parameters):
        self.extra_parameters = {}
        for param in self.algorithm().parameterDefinitions():
            if param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden:
                if param.name() in parameters:
                    self.extra_parameters[param.name()] = parameters[param.name()]
                continue

            if not param.name() in parameters:
                continue

            value = parameters[param.name()]

            wrapper = self.wrappers[param.name()]
            wrapper.setParameterValue(value, self.processing_context)

    def parameterChanged(self):
        """
        Called when a parameter value is changed in the panel
        """
        wrapper: QgsAbstractProcessingParameterWidgetWrapper = self.sender()
        default_values = self.algorithm().autogenerateParameterValues(
            self.createProcessingParameters(
                QgsProcessingParametersGenerator.Flag.SkipValidation
            ),
            wrapper.parameterDefinition().name(),
            Qgis.ProcessingMode.Standard,
        )
        if default_values:
            self.setParameters(default_values)



"""
***************************************************************************
    py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import math
import os
import sys
import time
import uuid
from typing import Optional

from qgis.core import QgsApplication, QgsProcessingContext, QgsProcessingUtils
from qgis.PyQt.QtCore import QDir

numExported = 1


def userFolder():
    userDir = os.path.join(QgsApplication.qgisSettingsDirPath(), "processing")
    if not QDir(userDir).exists():
        QDir().mkpath(userDir)

    return str(QDir.toNativeSeparators(userDir))


def defaultOutputFolder():
    folder = os.path.join(QDir.homePath(), "processing")
    return str(QDir.toNativeSeparators(folder))


def isWindows():
    return os.name == "nt"


def isMac():
    return sys.platform == "darwin"


def getTempFilename(ext=None, context: Optional[QgsProcessingContext] = None):
    tmpPath = QgsProcessingUtils.tempFolder(context)
    t = time.time()
    m = math.floor(t)
    uid = f"{m:8x}{int((t - m) * 1000000):05x}"
    if ext is None:
        filename = os.path.join(tmpPath, f"{uid}{getNumExportedLayers()}")
    else:
        filename = os.path.join(tmpPath, f"{uid}{getNumExportedLayers()}.{ext}")
    return filename


def getNumExportedLayers():
    global numExported
    numExported += 1
    return numExported


def mkdir(newdir):
    os.makedirs(newdir.strip("\n\r "), exist_ok=True)


def tempHelpFolder():
    tmp = os.path.join(str(QDir.tempPath()), "processing_help")
    if not QDir(tmp).exists():
        QDir().mkpath(tmp)

    return str(os.path.abspath(tmp))


def escapeAndJoin(strList):
    """
    .. deprecated:: 3.0
    Do not use, will be removed in QGIS 5.0
    """

    from warnings import warn

    warn(
        "processing.escapeAndJoin is deprecated and will be removed in QGIS 5.0",
        DeprecationWarning,
    )

    joined = ""
    for s in strList:
        if s[0] != "-" and " " in s:
            escaped = '"' + s.replace("\\", "\\\\").replace('"', '\\"') + '"'
        else:
            escaped = s
        joined += escaped + " "
    return joined.strip()



"""
***************************************************************************
    contour.py
    ---------------------
    Date                 : September 2013
    Copyright            : (C) 2013 by Alexander Bruy
    Email                : alexander bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "September 2013"
__copyright__ = "(C) 2013, Alexander Bruy"

import os

from qgis.core import (
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class contour(GdalAlgorithm):
    INPUT = "INPUT"
    BAND = "BAND"
    INTERVAL = "INTERVAL"
    FIELD_NAME = "FIELD_NAME"
    CREATE_3D = "CREATE_3D"
    IGNORE_NODATA = "IGNORE_NODATA"
    NODATA = "NODATA"
    OFFSET = "OFFSET"
    EXTRA = "EXTRA"
    OPTIONS = "OPTIONS"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBand(
                self.BAND,
                self.tr("Band number"),
                1,
                parentLayerParameterName=self.INPUT,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.INTERVAL,
                self.tr("Interval between contour lines"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                defaultValue=10.0,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.FIELD_NAME,
                self.tr(
                    "Attribute name (if not set, no elevation attribute is attached)"
                ),
                defaultValue="ELEV",
                optional=True,
            )
        )

        create_3d_param = QgsProcessingParameterBoolean(
            self.CREATE_3D, self.tr("Produce 3D vector"), defaultValue=False
        )
        create_3d_param.setFlags(
            create_3d_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(create_3d_param)

        ignore_nodata_param = QgsProcessingParameterBoolean(
            self.IGNORE_NODATA,
            self.tr("Treat all raster values as valid"),
            defaultValue=False,
        )
        ignore_nodata_param.setFlags(
            ignore_nodata_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(ignore_nodata_param)

        nodata_param = QgsProcessingParameterNumber(
            self.NODATA,
            self.tr("Input pixel value to treat as NoData"),
            type=QgsProcessingParameterNumber.Type.Double,
            defaultValue=None,
            optional=True,
        )
        nodata_param.setFlags(
            nodata_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(nodata_param)

        offset_param = QgsProcessingParameterNumber(
            self.OFFSET,
            self.tr("Offset from zero relative to which to interpret intervals"),
            type=QgsProcessingParameterNumber.Type.Double,
            defaultValue=0.0,
            optional=True,
        )
        nodata_param.setFlags(
            offset_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(offset_param)

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        # TODO: remove in QGIS 5
        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.FlagHidden
        )
        self.addParameter(options_param)

        self.addParameter(
            QgsProcessingParameterVectorDestination(
                self.OUTPUT,
                self.tr("Contours"),
                QgsProcessing.SourceType.TypeVectorLine,
            )
        )

    def name(self):
        return "contour"

    def displayName(self):
        return self.tr("Contour")

    def icon(self):
        return QIcon(os.path.join(pluginPath, "images", "gdaltools", "contour.png"))

    def group(self):
        return self.tr("Raster extraction")

    def groupId(self):
        return "rasterextraction"

    def commandName(self):
        return "gdal_contour"

    def _buildArgsList(self, parameters, context, feedback, executing):
        inLayer = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if inLayer is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )
        input_details = GdalUtils.gdal_connection_details_from_layer(inLayer)

        fieldName = self.parameterAsString(parameters, self.FIELD_NAME, context)
        if self.NODATA in parameters and parameters[self.NODATA] is not None:
            nodata = self.parameterAsDouble(parameters, self.NODATA, context)
        else:
            nodata = None
        offset = self.parameterAsDouble(parameters, self.OFFSET, context)

        outFile = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, outFile)
        output_details = GdalUtils.gdal_connection_details_from_uri(outFile, context)

        arguments = ["-b", str(self.parameterAsInt(parameters, self.BAND, context))]
        if fieldName:
            arguments.append("-a")
            arguments.append(fieldName)

        arguments.append("-i")
        arguments.append(
            str(self.parameterAsDouble(parameters, self.INTERVAL, context))
        )

        if self.parameterAsBoolean(parameters, self.CREATE_3D, context):
            arguments.append("-3d")

        if self.parameterAsBoolean(parameters, self.IGNORE_NODATA, context):
            arguments.append("-inodata")

        if nodata is not None:
            arguments.append(f"-snodata {nodata}")

        if offset:
            arguments.append(f"-off {offset}")

        if output_details.format:
            arguments.append(f"-f {output_details.format}")

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        # TODO: remove in QGIS 5
        options = self.parameterAsString(parameters, self.OPTIONS, context)
        if options:
            arguments.append(options)

        arguments.append(input_details.connection_string)
        arguments.append(output_details.connection_string)
        return arguments

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        arguments = self._buildArgsList(parameters, context, feedback, executing)
        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]


class contour_polygon(contour):
    FIELD_NAME_MIN = "FIELD_NAME_MIN"
    FIELD_NAME_MAX = "FIELD_NAME_MAX"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        super().initAlgorithm(config)
        # FIELD_NAME isn't used in polygon mode
        self.removeParameter(contour.FIELD_NAME)

        self.addParameter(
            QgsProcessingParameterString(
                self.FIELD_NAME_MIN,
                self.tr("Attribute name for minimum elevation of contour polygon"),
                defaultValue="ELEV_MIN",
                optional=True,
            )
        )

        self.addParameter(
            QgsProcessingParameterString(
                self.FIELD_NAME_MAX,
                self.tr("Attribute name for maximum elevation of contour polygon"),
                defaultValue="ELEV_MAX",
                optional=True,
            )
        )

        # Need to replace the output parameter, as we are producing a different type of output
        self.removeParameter(contour.OUTPUT)
        self.addParameter(
            QgsProcessingParameterVectorDestination(
                contour.OUTPUT,
                self.tr("Contours"),
                QgsProcessing.SourceType.TypeVectorPolygon,
            )
        )

    def name(self):
        return "contour_polygon"

    def displayName(self):
        return self.tr("Contour Polygons")

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        arguments = self._buildArgsList(parameters, context, feedback, executing)

        fieldNameMin = self.parameterAsString(parameters, self.FIELD_NAME_MIN, context)
        fieldNameMax = self.parameterAsString(parameters, self.FIELD_NAME_MAX, context)

        if fieldNameMin:
            arguments.insert(0, fieldNameMin)
            arguments.insert(0, "-amin")

        if fieldNameMax:
            arguments.insert(0, fieldNameMax)
            arguments.insert(0, "-amax")

        arguments.insert(0, "-p")

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]



"""
***************************************************************************
    QtSvgWidgets.py
    ---------------------
    Date                 : June 2026
    Copyright            : (C) 2026 Jean Felder
    Email                : jean dot felder at oslandia dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = 'Jean Felder'
__date__ = 'June 2026'
__copyright__ = '(C) 2026, Jean Felder'

from PyQt6.QtSvgWidgets import *



"""
***************************************************************************
    __init__.py
    ---------------------
    Date                 : January 2016
    Copyright            : (C) 2016 by Matthias Kuhn
    Email                : matthias@opengis.ch
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Matthias Kuhn"
__date__ = "January 2016"
__copyright__ = "(C) 2016, Matthias Kuhn"

import difflib
import filecmp
import functools
import hashlib
import inspect
import os
import sys
import tempfile
import unittest
from pathlib import Path
from typing import Optional, Union
from warnings import warn

from qgis.core import (
    NULL,
    QgsApplication,
    QgsCoordinateReferenceSystem,
    QgsFeatureRequest,
    QgsLayout,
    QgsLayoutChecker,
    QgsMapSettings,
    QgsMultiRenderChecker,
    QgsRenderChecker,
    QgsVectorLayer,
)
from qgis.PyQt.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QDir,
    QSize,
    Qt,
    QUrl,
    QVariant,
)
from qgis.PyQt.QtGui import QDesktopServices, QImage, QPainter

unittest.util._MAX_LENGTH = 2000


class QgisTestCase(unittest.TestCase):
    @staticmethod
    def is_ci_run() -> bool:
        """
        Returns True if the test is being run on the CI environment
        """
        return os.environ.get("QGIS_CONTINUOUS_INTEGRATION_RUN") == "true"

    @classmethod
    def setUpClass(cls):
        cls.report = ""
        cls.markdown_report = ""

    @classmethod
    def tearDownClass(cls):
        if cls.report:
            cls.write_local_html_report(cls.report)
        if cls.markdown_report:
            cls.write_local_markdown_report(cls.markdown_report)

    @classmethod
    def control_path_prefix(cls) -> Optional[str]:
        """
        Returns the prefix for test control images used by the class
        """
        return None

    @classmethod
    def write_local_html_report(cls, report: str):
        report_dir = QgsRenderChecker.testReportDir()
        if not report_dir.exists():
            QDir().mkpath(report_dir.path())

        report_file = report_dir.filePath("index.html")

        # only append to existing reports if running under CI
        file_is_empty = True
        if cls.is_ci_run() or os.environ.get("QGIS_APPEND_TO_TEST_REPORT") == "true":
            file_mode = "ta"
            try:
                with open(report_file, encoding="utf-8") as f:
                    file_is_empty = not bool(f.read())
            except OSError:
                pass
        else:
            file_mode = "wt"

        with open(report_file, file_mode, encoding="utf-8") as f:
            if file_is_empty:
                from .test_data_dir import TEST_DATA_DIR

                # append standard header
                with open(
                    TEST_DATA_DIR + "/../test_report_header.html", encoding="utf-8"
                ) as header_file:
                    f.write(header_file.read())

                # append embedded scripts
                f.write("<script>\n")
                with open(
                    TEST_DATA_DIR + "/../renderchecker.js", encoding="utf-8"
                ) as script_file:
                    f.write(script_file.read())
                f.write("</script>\n")

            f.write(f"<h1>Python {cls.__name__} Tests</h1>\n")
            f.write(report)

        if not QgisTestCase.is_ci_run():
            QDesktopServices.openUrl(QUrl.fromLocalFile(report_file))

    @classmethod
    def write_local_markdown_report(cls, report: str):
        report_dir = QgsRenderChecker.testReportDir()
        if not report_dir.exists():
            QDir().mkpath(report_dir.path())

        report_file = report_dir.filePath("summary.md")

        # only append to existing reports if running under CI
        if cls.is_ci_run() or os.environ.get("QGIS_APPEND_TO_TEST_REPORT") == "true":
            file_mode = "ta"
        else:
            file_mode = "wt"

        with open(report_file, file_mode, encoding="utf-8") as f:
            f.write(report)

    @classmethod
    def get_test_caller_details(
        cls,
    ) -> tuple[Optional[str], Optional[str], Optional[int]]:
        """
        Retrieves the details of the caller at the earliest position
        in the stack, excluding unittest internals.

        Returns the file, function name and line number of the caller
        """
        test_caller = None
        for caller_frame_record in inspect.stack():
            frame = caller_frame_record[0]
            info = inspect.getframeinfo(frame)
            # we want the highest level caller which isn't from unittest
            if "unittest" in info.filename or info.function == "_callTestMethod":
                break
            else:
                test_caller = info

        if test_caller:
            return (test_caller.filename, test_caller.function, test_caller.lineno)
        return None, None, None

    @classmethod
    def image_check(
        cls,
        name: str,
        reference_image: str,
        image: QImage,
        control_name=None,
        color_tolerance: int = 2,
        allowed_mismatch: int = 20,
        size_tolerance: Optional[Union[int, QSize]] = None,
        expect_fail: bool = False,
        control_path_prefix: Optional[str] = None,
        use_checkerboard_background: bool = False,
    ) -> bool:
        if use_checkerboard_background:
            output_image = QImage(image.size(), QImage.Format.Format_RGB32)
            QgsMultiRenderChecker.drawBackground(output_image)
            painter = QPainter(output_image)
            painter.drawImage(0, 0, image)
            painter.end()
            image = output_image

        temp_dir = QDir.tempPath() + "/"
        file_name = temp_dir + name + ".png"
        image.save(file_name, "PNG")
        checker = QgsMultiRenderChecker()

        caller_file, caller_function, caller_line_no = cls.get_test_caller_details()
        if caller_file:
            checker.setFileFunctionLine(caller_file, caller_function, caller_line_no)

        if control_path_prefix:
            checker.setControlPathPrefix(control_path_prefix)
        elif cls.control_path_prefix():
            checker.setControlPathPrefix(cls.control_path_prefix())

        checker.setControlName(control_name or "expected_" + reference_image)
        checker.setRenderedImage(file_name)
        checker.setColorTolerance(color_tolerance)
        checker.setExpectFail(expect_fail)
        if size_tolerance is not None:
            if isinstance(size_tolerance, QSize):
                if size_tolerance.isValid():
                    checker.setSizeTolerance(
                        size_tolerance.width(), size_tolerance.height()
                    )
            else:
                checker.setSizeTolerance(size_tolerance, size_tolerance)

        result = checker.runTest(name, allowed_mismatch)
        if (not expect_fail and not result) or (expect_fail and result):
            cls.report += f"<h2>Render {name}</h2>\n"
            cls.report += checker.report()

            markdown = checker.markdownReport()
            if markdown:
                cls.markdown_report += f"## {name}\n\n"
                cls.markdown_report += markdown

        return result

    @classmethod
    def render_map_settings_check(
        cls,
        name: str,
        reference_image: str,
        map_settings: QgsMapSettings,
        control_name=None,
        color_tolerance: Optional[int] = None,
        allowed_mismatch: Optional[int] = None,
        control_path_prefix: Optional[str] = None,
    ) -> bool:
        checker = QgsMultiRenderChecker()
        checker.setMapSettings(map_settings)

        caller_file, caller_function, caller_line_no = cls.get_test_caller_details()
        if caller_file:
            checker.setFileFunctionLine(caller_file, caller_function, caller_line_no)

        if control_path_prefix:
            checker.setControlPathPrefix(control_path_prefix)
        elif cls.control_path_prefix():
            checker.setControlPathPrefix(cls.control_path_prefix())
        checker.setControlName(control_name or "expected_" + reference_image)
        if color_tolerance:
            checker.setColorTolerance(color_tolerance)
        result = checker.runTest(name, allowed_mismatch or 0)
        if not result:
            cls.report += f"<h2>Render {name}</h2>\n"
            cls.report += checker.report()

            markdown = checker.markdownReport()
            if markdown:
                cls.markdown_report += f"## {name}\n\n"
                cls.markdown_report += markdown

        return result

    @classmethod
    def render_layout_check(
        cls,
        name: str,
        layout: QgsLayout,
        size: Optional[QSize] = None,
        color_tolerance: Optional[int] = None,
        allowed_mismatch: Optional[int] = None,
        page: Optional[int] = 0,
    ) -> bool:
        checker = QgsLayoutChecker(name, layout)

        caller_file, caller_function, caller_line_no = cls.get_test_caller_details()
        if caller_file:
            checker.setFileFunctionLine(caller_file, caller_function, caller_line_no)

        if size is not None:
            checker.setSize(size)
        if color_tolerance is not None:
            checker.setColorTolerance(color_tolerance)

        if cls.control_path_prefix():
            checker.setControlPathPrefix(cls.control_path_prefix())
        result, message = checker.testLayout(page=page, pixelDiff=allowed_mismatch or 0)
        if not result:
            cls.report += f"<h2>Render {name}</h2>\n"
            cls.report += checker.report()

            markdown = checker.markdownReport()
            if markdown:
                cls.markdown_report += f"## {name}\n\n"
                cls.markdown_report += markdown

        return result

    @staticmethod
    def get_test_data_path(file_path: str) -> Path:
        """
        Returns the full path to a file contained within the test data
        directory.
        """
        from utilities import unitTestDataPath

        return Path(unitTestDataPath()) / (
            file_path[1:] if file_path.startswith("/") else file_path
        )

    @staticmethod
    def strip_std_ignorable_errors(output: str) -> str:
        """
        Strips out ignorable warnings and errors from stdout/stderr output
        """
        return "\n".join(
            [
                e
                for e in output.splitlines()
                if e
                not in (
                    "Problem with GRASS installation: GRASS was not found or is not correctly installed",
                    "QStandardPaths: wrong permissions on runtime directory /tmp, 0777 instead of 0700",
                    "MESA: error: ZINK: failed to choose pdev",
                    "MESA: error: ZINK: vkEnumeratePhysicalDevices failed (VK_ERROR_INITIALIZATION_FAILED)",
                    "glx: failed to create drisw screen",
                    "failed to load driver: zink",
                    "QML debugging is enabled. Only use this in a safe environment.",
                    "<frozen importlib._bootstrap>:488: DeprecationWarning: builtin type SwigPyObject has no __module__ attribute",
                    "<frozen importlib._bootstrap>:488: DeprecationWarning: builtin type SwigPyPacked has no __module__ attribute",
                )
                and not "LC_ALL: cannot change locale" in e
            ]
        )

    @staticmethod
    def sanitize_local_url(endpoint: str, query: str) -> str:
        """
        Sanitizes a URL to a local file path. Matches logic in c++ code so that
        remote URLs can be mocked to local test files.
        """
        if not os.path.exists(endpoint):
            os.makedirs(endpoint)

        if query.startswith("/query"):
            query = query[len("/query") :]
            endpoint = endpoint + "_query"

        if query.startswith("/exportImage"):
            query = query[len("/exportImage") :]
            endpoint = endpoint + "_exportImage"

        if len(endpoint + query) > 150:
            ret = endpoint + hashlib.md5(query.encode()).hexdigest()
            # print('Before: ' + endpoint + query)
            # print('After:  ' + ret)
            return ret
        return endpoint + query.replace("?", "_").replace("&", "_").replace(
            "<", "_"
        ).replace(">", "_").replace('"', "_").replace("'", "_").replace(
            " ", "_"
        ).replace(":", "_").replace("/", "_").replace("\n", "_")

    def assertLayersEqual(self, layer_expected, layer_result, **kwargs):
        """
        :param layer_expected: The first layer to compare
        :param layer_result: The second layer to compare
        :param request: Optional, A feature request. This can be used to specify
                        an order by clause to make sure features are compared in
                        a given sequence if they don't match by default.
        :keyword compare: A map of comparison options. e.g.
                         { fields: { a: skip, b: { precision: 2 }, geometry: { precision: 5 } }
                         { fields: { __all__: cast( str ) } }
        :keyword pk: "Primary key" type field - used to match features
        from the expected table to their corresponding features in the result table. If not specified
        features are compared by their order in the layer (e.g. first feature compared with first feature,
        etc)
        """
        self.checkLayersEqual(layer_expected, layer_result, True, **kwargs)

    def checkLayersEqual(
        self, layer_expected, layer_result, use_asserts=False, **kwargs
    ):
        """
        :param layer_expected: The first layer to compare
        :param layer_result: The second layer to compare
        :param use_asserts: If true, asserts are used to test conditions, if false, asserts
        are not used and the function will only return False if the test fails
        :param request: Optional, A feature request. This can be used to specify
                        an order by clause to make sure features are compared in
                        a given sequence if they don't match by default.
        :keyword compare: A map of comparison options. e.g.
                         { fields: { a: skip, b: { precision: 2 }, geometry: { precision: 5 } }
                         { fields: { __all__: cast( str ) } }
        :keyword pk: "Primary key" type field - used to match features
        from the expected table to their corresponding features in the result table. If not specified
        features are compared by their order in the layer (e.g. first feature compared with first feature,
        etc)
        """

        try:
            request = kwargs["request"]
        except KeyError:
            request = QgsFeatureRequest()

        try:
            compare = kwargs["compare"]
        except KeyError:
            compare = {}

        # Compare CRS
        if "ignore_crs_check" not in compare or not compare["ignore_crs_check"]:
            expected_wkt = (
                layer_expected.dataProvider()
                .crs()
                .toWkt(QgsCoordinateReferenceSystem.WktVariant.WKT_PREFERRED)
            )
            result_wkt = (
                layer_result.dataProvider()
                .crs()
                .toWkt(QgsCoordinateReferenceSystem.WktVariant.WKT_PREFERRED)
            )

            if use_asserts:
                self.assertEqual(
                    layer_expected.dataProvider().crs(),
                    layer_result.dataProvider().crs(),
                )
            elif (
                layer_expected.dataProvider().crs() != layer_result.dataProvider().crs()
            ):
                return False

        # Compare features
        if use_asserts:
            self.assertEqual(layer_expected.featureCount(), layer_result.featureCount())
        elif layer_expected.featureCount() != layer_result.featureCount():
            return False

        try:
            precision = compare["geometry"]["precision"]
        except KeyError:
            precision = 14

        try:
            topo_equal_check = compare["geometry"]["topo_equal_check"]
        except KeyError:
            topo_equal_check = False

        try:
            ignore_part_order = compare["geometry"]["ignore_part_order"]
        except KeyError:
            ignore_part_order = False

        try:
            normalize = compare["geometry"]["normalize"]
        except KeyError:
            normalize = False

        try:
            explode_collections = compare["geometry"]["explode_collections"]
        except KeyError:
            explode_collections = False

        try:
            snap_to_grid = compare["geometry"]["snap_to_grid"]
        except KeyError:
            snap_to_grid = None

        try:
            unordered = compare["unordered"]
        except KeyError:
            unordered = False

        try:
            equate_null_and_empty = compare["geometry"]["equate_null_and_empty"]
        except KeyError:
            equate_null_and_empty = False

        if unordered:
            features_expected = [f for f in layer_expected.getFeatures(request)]
            for feat in layer_result.getFeatures(request):
                feat_expected_equal = None
                for feat_expected in features_expected:
                    if self.checkGeometriesEqual(
                        feat.geometry(),
                        feat_expected.geometry(),
                        feat.id(),
                        feat_expected.id(),
                        False,
                        precision,
                        topo_equal_check,
                        ignore_part_order,
                        normalize=normalize,
                        explode_collections=explode_collections,
                        snap_to_grid=snap_to_grid,
                        equate_null_and_empty=equate_null_and_empty,
                    ) and self.checkAttributesEqual(
                        feat, feat_expected, layer_expected.fields(), False, compare
                    ):
                        feat_expected_equal = feat_expected
                        break

                if feat_expected_equal is not None:
                    features_expected.remove(feat_expected_equal)
                else:
                    if use_asserts:
                        self.assertTrue(
                            False,
                            "Unexpected result feature: fid {}, geometry: {}, attributes: {}".format(
                                feat.id(),
                                (
                                    feat.geometry().constGet().asWkt(precision)
                                    if feat.geometry()
                                    else "NULL"
                                ),
                                feat.attributes(),
                            ),
                        )
                    else:
                        return False

            if len(features_expected) != 0:
                if use_asserts:
                    lst_missing = []
                    for feat in features_expected:
                        lst_missing.append(
                            "fid {}, geometry: {}, attributes: {}".format(
                                feat.id(),
                                (
                                    feat.geometry().constGet().asWkt(precision)
                                    if feat.geometry()
                                    else "NULL"
                                ),
                                feat.attributes(),
                            )
                        )
                    self.assertTrue(
                        False,
                        "Some expected features not found in results:\n"
                        + "\n".join(lst_missing),
                    )
                else:
                    return False

            return True

        def get_pk_or_fid(f):
            if "pk" in kwargs and kwargs["pk"] is not None:
                key = kwargs["pk"]
                if isinstance(key, list) or isinstance(key, tuple):
                    return [f[k] for k in key]
                else:
                    return f[kwargs["pk"]]
            else:
                return f.id()

        def sort_by_pk_or_fid(f):
            pk = get_pk_or_fid(f)
            # we want NULL values sorted first, and don't want to try to
            # directly compare NULL against non-NULL values
            if isinstance(pk, list):
                pk = [(v == NULL, v) for v in pk]
            return (pk == NULL, pk)

        expected_features = sorted(
            layer_expected.getFeatures(request), key=sort_by_pk_or_fid
        )
        result_features = sorted(
            layer_result.getFeatures(request), key=sort_by_pk_or_fid
        )

        for feats in zip(expected_features, result_features):
            eq = self.checkGeometriesEqual(
                feats[0].geometry(),
                feats[1].geometry(),
                feats[0].id(),
                feats[1].id(),
                use_asserts,
                precision,
                topo_equal_check,
                ignore_part_order,
                normalize=normalize,
                explode_collections=explode_collections,
                snap_to_grid=snap_to_grid,
                equate_null_and_empty=equate_null_and_empty,
            )
            if not eq and not use_asserts:
                return False

            eq = self.checkAttributesEqual(
                feats[0], feats[1], layer_expected.fields(), use_asserts, compare
            )
            if not eq and not use_asserts:
                return False

        return True

    def checkFilesEqual(self, filepath_expected, filepath_result, use_asserts=False):
        with open(filepath_expected) as file_expected:
            with open(filepath_result) as file_result:
                diff = difflib.unified_diff(
                    file_expected.readlines(),
                    file_result.readlines(),
                    fromfile="expected",
                    tofile="result",
                )
                diff = list(diff)
                eq = not len(diff)
                if use_asserts:
                    self.assertEqual(0, len(diff), "".join(diff))
                else:
                    return eq

    def assertFilesEqual(self, filepath_expected, filepath_result):
        self.checkFilesEqual(filepath_expected, filepath_result, use_asserts=True)

    def assertDirectoryEqual(self, dirpath_expected: str, dirpath_result: str):
        """
        Checks whether both directories have the same content (non-recursively) and raises an assertion error if not.
        """
        path_expected = Path(dirpath_expected)
        path_result = Path(dirpath_result)

        contents_result = list(path_result.iterdir())
        contents_expected = list(path_expected.iterdir())
        contents_expected = [
            p
            for p in contents_expected
            if p.suffix != ".png" or not p.stem.endswith("_mask")
        ]
        self.assertCountEqual(
            [p.name if p.is_file() else p.stem for p in contents_expected],
            [p.name if p.is_file() else p.stem for p in contents_result],
            f"Directory contents mismatch in {dirpath_expected} vs {dirpath_result}",
        )

        # compare file contents
        for expected_file_path in contents_expected:
            if expected_file_path.is_dir():
                continue

            result_file_path = path_result / expected_file_path.name

            if expected_file_path.suffix == ".pbf":
                # vector layer, use assertLayersEqual
                layer_expected = QgsVectorLayer(str(expected_file_path), "Expected")
                self.assertTrue(layer_expected.isValid())
                layer_result = QgsVectorLayer(str(result_file_path), "Result")
                self.assertTrue(layer_result.isValid())
                self.assertLayersEqual(layer_expected, layer_result)
            elif expected_file_path.suffix == ".png":
                # image file, use QgsRenderChecker
                checker = QgsRenderChecker()
                res = checker.compareImages(
                    expected_file_path.stem,
                    expected_file_path.as_posix(),
                    result_file_path.as_posix(),
                )
                self.assertTrue(res)
            else:
                assert False, (
                    f"Don't know how to compare {expected_file_path.suffix} files"
                )

    def assertDirectoriesEqual(self, dirpath_expected: str, dirpath_result: str):
        """Checks whether both directories have the same content (recursively) and raises an assertion error if not."""
        self.assertDirectoryEqual(dirpath_expected, dirpath_result)

        # recurse through subfolders
        path_expected = Path(dirpath_expected)
        path_result = Path(dirpath_result)
        for p in path_expected.iterdir():
            if p.is_dir():
                self.assertDirectoriesEqual(str(p), path_result / p.stem)

    def assertGeometriesEqual(
        self,
        geom0,
        geom1,
        geom0_id="geometry 1",
        geom1_id="geometry 2",
        precision=14,
        topo_equal_check=False,
        ignore_part_order=False,
        normalize=False,
        explode_collections=False,
        snap_to_grid=None,
        equate_null_and_empty=False,
    ):
        self.checkGeometriesEqual(
            geom0,
            geom1,
            geom0_id,
            geom1_id,
            use_asserts=True,
            precision=precision,
            topo_equal_check=topo_equal_check,
            ignore_part_order=ignore_part_order,
            normalize=normalize,
            explode_collections=explode_collections,
            snap_to_grid=snap_to_grid,
            equate_null_and_empty=equate_null_and_empty,
        )

    def checkGeometriesEqual(
        self,
        geom0,
        geom1,
        geom0_id,
        geom1_id,
        use_asserts=False,
        precision=14,
        topo_equal_check=False,
        ignore_part_order=False,
        normalize=False,
        explode_collections=False,
        snap_to_grid=None,
        equate_null_and_empty=False,
    ):
        """Checks whether two geometries are the same - using either a strict check of coordinates (up to given precision)
        or by using topological equality (where e.g. a polygon with clockwise is equal to a polygon with counter-clockwise
        order of vertices)
        .. versionadded:: 3.2
        """
        geom0_wkt = ""
        geom0_wkt_full = ""
        geom1_wkt = ""
        geom1_wkt_full = ""

        geom0_is_null = geom0.isNull() or (equate_null_and_empty and geom0.isEmpty())
        geom1_is_null = geom1.isNull() or (equate_null_and_empty and geom1.isEmpty())
        if not geom0_is_null and not geom1_is_null:
            if snap_to_grid is not None:
                geom0 = geom0.snappedToGrid(
                    snap_to_grid, snap_to_grid, snap_to_grid, snap_to_grid
                )
                geom1 = geom1.snappedToGrid(
                    snap_to_grid, snap_to_grid, snap_to_grid, snap_to_grid
                )
            if normalize:
                geom0.normalize()
                geom1.normalize()

            raw_geom0 = geom0.constGet()
            raw_geom1 = geom1.constGet()
            if explode_collections:
                raw_geom0 = raw_geom0.simplifiedTypeRef()
                raw_geom1 = raw_geom1.simplifiedTypeRef()
            geom0_wkt = raw_geom0.asWkt(precision)
            geom0_wkt_full = raw_geom0.asWkt()
            geom1_wkt = raw_geom1.asWkt(precision)
            geom1_wkt_full = raw_geom1.asWkt()
            equal = geom0_wkt == geom1_wkt
            if not equal and topo_equal_check:
                equal = geom0.isGeosEqual(geom1)
            if not equal and ignore_part_order and geom0.isMultipart():
                equal = sorted(
                    [p.asWkt(precision) for p in geom0.constParts()]
                ) == sorted([p.asWkt(precision) for p in geom1.constParts()])
        elif geom0_is_null and geom1_is_null:
            equal = True
        else:
            geom0_wkt = geom0.asWkt(precision)
            geom1_wkt = geom1.asWkt(precision)
            geom0_wkt_full = geom0.asWkt()
            geom1_wkt_full = geom1.asWkt()
            equal = False

        if use_asserts:
            self.assertTrue(
                equal,
                ""
                " Features (Expected fid: {}, Result fid: {}) differ in geometry with method {}: \n\n"
                "  {}\n"
                "  At given precision ({}):\n"
                "   Expected geometry: {}\n"
                "   Result geometry:   {}\n\n"
                "  Full precision:\n"
                "   Expected geometry : {}\n"
                "   Result geometry:   {}\n\n".format(
                    geom0_id,
                    geom1_id,
                    "geos" if topo_equal_check else "wkt",
                    "Normalized" if normalize else "Not-normalized",
                    precision,
                    geom0_wkt if not geom0_is_null else "NULL",
                    geom1_wkt if not geom1_is_null else "NULL",
                    geom0_wkt_full if not geom0_is_null else "NULL",
                    geom1_wkt_full if not geom1_is_null else "NULL",
                ),
            )
        else:
            return equal

    def checkAttributesEqual(self, feat0, feat1, fields_expected, use_asserts, compare):
        """Checks whether attributes of two features are the same
        .. versionadded:: 3.2
        """

        for attr_expected, field_expected in zip(
            feat0.attributes(), fields_expected.toList()
        ):
            try:
                cmp = compare["fields"][field_expected.name()]
            except KeyError:
                try:
                    cmp = compare["fields"]["__all__"]
                except KeyError:
                    cmp = {}

            # Skip field
            if "skip" in cmp:
                continue

            if use_asserts:
                self.assertIn(
                    field_expected.name().lower(),
                    [name.lower() for name in feat1.fields().names()],
                )

            attr_result = feat1[field_expected.name()]
            field_result = [
                fld
                for fld in fields_expected.toList()
                if fld.name() == field_expected.name()
            ][0]

            # Cast field to a given type
            isNumber = False
            if "cast" in cmp:
                if cmp["cast"] == "int":
                    attr_expected = int(attr_expected) if attr_expected else None
                    attr_result = int(attr_result) if attr_result else None
                    isNumber = True
                if cmp["cast"] == "float":
                    attr_expected = float(attr_expected) if attr_expected else None
                    attr_result = float(attr_result) if attr_result else None
                    isNumber = True
                if cmp["cast"] == "str":
                    if isinstance(attr_expected, QDateTime):
                        attr_expected = attr_expected.toString("yyyy/MM/dd hh:mm:ss")
                    elif isinstance(attr_expected, QDate):
                        attr_expected = attr_expected.toString("yyyy/MM/dd")
                    else:
                        attr_expected = str(attr_expected) if attr_expected else None
                    if isinstance(attr_result, QDateTime):
                        attr_result = attr_result.toString("yyyy/MM/dd hh:mm:ss")
                    elif isinstance(attr_result, QDate):
                        attr_result = attr_result.toString("yyyy/MM/dd")
                    else:
                        attr_result = str(attr_result) if attr_result else None

            # Round field (only numeric so it works with __all__)
            if "precision" in cmp and (
                field_expected.type()
                in [QVariant.Int, QVariant.Double, QVariant.LongLong]
                or isNumber
            ):
                if not attr_expected == NULL:
                    attr_expected = round(attr_expected, cmp["precision"])
                if not attr_result == NULL:
                    attr_result = round(attr_result, cmp["precision"])

            if use_asserts:
                self.assertEqual(
                    attr_expected,
                    attr_result,
                    f"Features {feat0.id()}/{feat1.id()} differ in attributes\n\n * Field expected: {field_expected.name()} ({field_expected.typeName()})\n * result  : {field_result.name()} ({field_result.typeName()})\n\n * Expected: {repr(attr_expected)} != Result  : {repr(attr_result)}",
                )
            elif attr_expected != attr_result:
                return False

        return True


class _UnexpectedSuccess(Exception):
    """
    The test was supposed to fail, but it didn't!
    """

    pass


def expectedFailure(*args):
    """
    Will decorate a unittest function as an expectedFailure. A function
    flagged as expectedFailure will be succeed if it raises an exception.
    If it does not raise an exception, this will throw an
    `_UnexpectedSuccess` exception.

        @expectedFailure
        def my_test(self):
            self.assertTrue(False)

    The decorator also accepts a parameter to only expect a failure under
    certain conditions.

        @expectedFailure(time.localtime().tm_year < 2002)
        def my_test(self):
            self.assertTrue(qgisIsInvented())
    """
    if hasattr(args[0], "__call__"):
        # We got a function as parameter: assume usage like
        #   @expectedFailure
        #   def testfunction():
        func = args[0]

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception:
                pass
            else:
                raise _UnexpectedSuccess

        return wrapper
    else:
        # We got a function as parameter: assume usage like
        #   @expectedFailure(failsOnThisPlatform)
        #   def testfunction():
        condition = args[0]

        def realExpectedFailure(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if condition:
                    try:
                        func(*args, **kwargs)
                    except Exception:
                        pass
                    else:
                        raise _UnexpectedSuccess
                else:
                    func(*args, **kwargs)

            return wrapper

        return realExpectedFailure


QgisTestCase.expectedFailure = expectedFailure


def _deprecatedAssertLayersEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.assertLayersEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.assertLayersEqual(*args, **kwargs)


def _deprecatedCheckLayersEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.checkLayersEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.checkLayersEqual(*args, **kwargs)


def _deprecatedAssertFilesEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.assertFilesEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.assertFilesEqual(*args, **kwargs)


def _deprecatedCheckFilesEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.checkFilesEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.checkFilesEqual(*args, **kwargs)


def _deprecatedAssertDirectoryEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.assertDirectoryEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.assertDirectoryEqual(*args, **kwargs)


def _deprecatedAssertDirectoriesEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.assertDirectoriesEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.assertDirectoriesEqual(*args, **kwargs)


def _deprecatedAssertGeometriesEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.assertGeometriesEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.assertGeometriesEqual(*args, **kwargs)


def _deprecatedCheckGeometriesEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.checkGeometriesEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.checkGeometriesEqual(*args, **kwargs)


def _deprecatedCheckAttributesEqual(*args, **kwargs):
    warn(
        "unittest.TestCase.checkAttributesEqual is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    return QgisTestCase.checkAttributesEqual(*args, **kwargs)


def _deprecated_image_check(*args, **kwargs):
    warn(
        "unittest.TestCase.image_check is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    # Remove the first args element `self`  which we don't need for a @classmethod
    return QgisTestCase.image_check(*args[1:], **kwargs)


def _deprecated_render_map_settings_check(*args, **kwargs):
    warn(
        "unittest.TestCase.render_map_settings_check is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    # Remove the first args element `self`  which we don't need for a @classmethod
    return QgisTestCase.render_map_settings_check(*args[1:], **kwargs)


def _deprecated_render_layout_check(*args, **kwargs):
    warn(
        "unittest.TestCase.render_layout_check is deprecated and will be removed in the future. Port your tests to `qgis.testing.TestCase`",
        DeprecationWarning,
    )
    # Remove the first args element `self`  which we don't need for a @classmethod
    return QgisTestCase.render_layout_check(*args[1:], **kwargs)


TestCase = unittest.TestCase
TestCase.assertLayersEqual = _deprecatedAssertLayersEqual
TestCase.checkLayersEqual = _deprecatedCheckLayersEqual
TestCase.assertFilesEqual = _deprecatedAssertFilesEqual
TestCase.checkFilesEqual = _deprecatedCheckFilesEqual
TestCase.assertDirectoryEqual = _deprecatedAssertDirectoryEqual
TestCase.assertDirectoriesEqual = _deprecatedAssertDirectoriesEqual
TestCase.assertGeometriesEqual = _deprecatedAssertGeometriesEqual
TestCase.checkGeometriesEqual = _deprecatedCheckGeometriesEqual
TestCase.checkAttributesEqual = _deprecatedCheckAttributesEqual
TestCase.image_check = _deprecated_image_check
TestCase.render_map_settings_check = _deprecated_render_map_settings_check
TestCase.render_layout_check = _deprecated_render_layout_check


def start_app(cleanup=True):
    """
    Will start a QgsApplication and call all initialization code like
    registering the providers and other infrastructure. It will not load
    any plugins.

    You can always get the reference to a running app by calling `QgsApplication.instance()`.

    The initialization will only happen once, so it is safe to call this method repeatedly.

        Parameters
        ----------

        cleanup: Do cleanup on exit. Defaults to true.

        Returns
        -------
        QgsApplication

        A QgsApplication singleton
    """
    global QGISAPP

    try:
        QGISAPP
    except NameError:
        myGuiFlag = True  # All test will run qgis in gui mode

        try:
            sys.argv
        except AttributeError:
            sys.argv = [""]

        # In python3 we need to convert to a bytes object (or should
        # QgsApplication accept a QString instead of const char* ?)
        try:
            argvb = list(map(os.fsencode, sys.argv))
        except AttributeError:
            argvb = sys.argv

        QCoreApplication.setAttribute(
            Qt.ApplicationAttribute.AA_ShareOpenGLContexts, True
        )

        QCoreApplication.setOrganizationName("QGIS")
        QCoreApplication.setOrganizationDomain("qgis.org")
        QCoreApplication.setApplicationName("QGIS-TEST")

        tmpdir = tempfile.mkdtemp("", "QGIS-PythonTestConfigPath-")
        os.environ["QGIS_CUSTOM_CONFIG_PATH"] = tmpdir

        # Note: QGIS_PREFIX_PATH is evaluated in QgsApplication -
        # no need to mess with it here.
        QGISAPP = QgsApplication(argvb, myGuiFlag)

        QGISAPP.initQgis()
        print(QGISAPP.showSettings())

        def debug_log_message(message, tag, level):
            print(f"{tag}({level}): {message}")

        QgsApplication.instance().messageLog().messageReceived.connect(
            debug_log_message
        )

        if cleanup:
            import atexit
            import shutil

            @atexit.register
            def exitQgis():
                QGISAPP.exitQgis()
                shutil.rmtree(tmpdir)

    return QGISAPP


def stop_app():
    """
    Cleans up and exits QGIS
    """
    global QGISAPP

    QGISAPP.exitQgis()
    del QGISAPP



###############################################################################
#
# MetaSearch Catalog Client
#
# Copyright (C) 2014 Tom Kralidis (tomkralidis@gmail.com)
#
# This source is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This code is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
###############################################################################

WMSWMST_LINK_TYPES = [
    "WMS",
    "WMTS",
    "OGC:WMS",
    "OGC:WMTS",
    "OGC:WMS-1.1.1-http-get-map",
    "OGC:WMS-1.1.1-http-get-capabilities",
    "OGC:WMS-1.3.0-http-get-map",
    "OGC:WMS-1.3.0-http-get-capabilities",
    "urn:x-esri:specification:ServiceType:wms:url",
    "urn:x-esri:specification:ServiceType:Gmd:URL.wms",
    "OGC Web Map Service",
    "OGC Web Map Tile Service",
]

WFS_LINK_TYPES = [
    "WFS",
    "OGC:WFS",
    "OGC:WFS-1.0.0-http-get-capabilities",
    "OGC:WFS-1.1.0-http-get-capabilities",
    "urn:x-esri:specification:ServiceType:wfs:url",
    "urn:x-esri:specification:ServiceType:Gmd:URL.wfs",
    "OGC Web Feature Service",
]

WCS_LINK_TYPES = [
    "WCS",
    "OGC:WCS",
    "OGC:WCS-1.1.0-http-get-capabilities",
    "urn:x-esri:specification:ServiceType:wcs:url",
    "urn:x-esri:specification:ServiceType:Gmd:URL.wcs",
    "OGC Web Coverage Service",
]

AMS_LINK_TYPES = [
    "ESRI:ArcGIS:MapServer",
    "Esri REST: Map Service",
    "ESRI REST",
    "MapServer",
]

AFS_LINK_TYPES = [
    "ESRI:ArcGIS:FeatureServer",
    "Esri REST: Feature Service",
    "FeatureServer",
]

GIS_FILE_LINK_TYPES = ["FILE:GEO"]



"""
***************************************************************************
    grass_algorithms_raster_test_pt2.py
    -----------------------------
    Date                 : May 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : mederic dot ribreux at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "May 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

import os
import shutil
import tempfile

import AlgorithmsTestBase
import nose2
from grassprovider.grass_provider import GrassProvider
from grassprovider.grass_utils import GrassUtils
from qgis.core import QgsApplication, QgsProcessingContext, QgsProcessingFeedback
from qgis.testing import QgisTestCase, start_app

testDataPath = os.path.join(os.path.dirname(__file__), "testdata")


class TestGrassAlgorithmsRasterTest(QgisTestCase, AlgorithmsTestBase.AlgorithmsTest):
    @classmethod
    def setUpClass(cls):
        start_app()
        cls.provider = GrassProvider()
        QgsApplication.processingRegistry().addProvider(cls.provider)
        cls.cleanup_paths = []

        cls.temp_dir = tempfile.mkdtemp()
        cls.cleanup_paths.append(cls.temp_dir)

        assert GrassUtils.installedVersion()

    @classmethod
    def tearDownClass(cls):
        QgsApplication.processingRegistry().removeProvider(cls.provider)
        for path in cls.cleanup_paths:
            shutil.rmtree(path)

    def definition_file(self):
        return "grass_algorithms_raster_tests2.yaml"

    def testNeighbors(self):
        context = QgsProcessingContext()
        input_raster = os.path.join(
            testDataPath, "custom", "grass7", "float_raster.tif"
        )

        alg = QgsApplication.processingRegistry().createAlgorithmById(
            "grass:r.neighbors"
        )
        self.assertIsNotNone(alg)

        temp_file = os.path.join(self.temp_dir, "grass_output.tif")

        # Test an even integer for neighborhood size
        parameters = {
            "input": input_raster,
            "selection": None,
            "method": 0,
            "size": 4,
            "gauss": None,
            "quantile": "",
            "-c": False,
            "-a": False,
            "weight": "",
            "output": temp_file,
            "GRASS_REGION_PARAMETER": None,
            "GRASS_REGION_CELLSIZE_PARAMETER": 0,
            "GRASS_RASTER_FORMAT_OPT": "",
            "GRASS_RASTER_FORMAT_META": "",
        }

        ok, msg = alg.checkParameterValues(parameters, context)
        self.assertFalse(ok)


if __name__ == "__main__":
    nose2.main()



"""
***************************************************************************
    r_statistics.py
    ---------------
    Date                 : September 2017
    Copyright            : (C) 2017 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "September 2017"
__copyright__ = "(C) 2017, Médéric Ribreux"

from qgis.core import QgsProcessingParameterString

from grassprovider.grass_utils import GrassUtils


def processCommand(alg, parameters, context, feedback):
    # We had a new "output" parameter
    out = f"output{alg.uniqueSuffix}"
    p = QgsProcessingParameterString("~output", None, out, False, False)
    alg.addParameter(p)

    # We need to remove all outputs
    alg.processCommand(parameters, context, feedback, True)

    # Then we add a new command for treating results
    calcExpression = f"correctedoutput{alg.uniqueSuffix}=@{out}"
    command = f'r.mapcalc expression="{calcExpression}"'
    alg.commands.append(command)


def processOutputs(alg, parameters, context, feedback):
    createOpt = alg.parameterAsString(parameters, alg.GRASS_RASTER_FORMAT_OPT, context)
    metaOpt = alg.parameterAsString(parameters, alg.GRASS_RASTER_FORMAT_META, context)

    # Export the results from correctedoutput
    grassName = f"correctedoutput{alg.uniqueSuffix}"
    fileName = alg.parameterAsOutputLayer(parameters, "routput", context)
    outFormat = GrassUtils.getRasterFormatFromFilename(fileName)
    alg.exportRasterLayer(grassName, fileName, True, outFormat, createOpt, metaOpt)



"""
***************************************************************************
    EliminateSelection.py
    ---------------------
    Date                 : January 2017
    Copyright         : (C) 2017 by Bernhard Ströbl
    Email                : bernhard.stroebl@jena.de
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Bernhard Ströbl"
__date__ = "January 2017"
__copyright__ = "(C) 2017, Bernhard Ströbl"

import os

from qgis.core import (
    QgsApplication,
    QgsFeature,
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsGeometry,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterVectorLayer,
    QgsProcessingUtils,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class EliminateSelection(QgisAlgorithm):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    MODE = "MODE"

    MODE_LARGEST_AREA = 0
    MODE_SMALLEST_AREA = 1
    MODE_BOUNDARY = 2

    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmDissolve.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmDissolve.svg")

    def group(self):
        return self.tr("Vector geometry")

    def groupId(self):
        return "vectorgeometry"

    def __init__(self):
        super().__init__()

    def flags(self):
        return (
            super().flags()
            | QgsProcessingAlgorithm.Flag.FlagNoThreading
            | QgsProcessingAlgorithm.Flag.FlagNotAvailableInStandaloneTool
        )

    def initAlgorithm(self, config=None):
        self.modes = [
            self.tr("Largest Area"),
            self.tr("Smallest Area"),
            self.tr("Largest Common Boundary"),
        ]

        self.addParameter(
            QgsProcessingParameterVectorLayer(
                self.INPUT,
                self.tr("Input layer"),
                [QgsProcessing.SourceType.TypeVectorPolygon],
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.MODE,
                self.tr("Merge selection with the neighbouring polygon with the"),
                options=self.modes,
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Eliminated"),
                QgsProcessing.SourceType.TypeVectorPolygon,
            )
        )

    def name(self):
        return "eliminateselectedpolygons"

    def displayName(self):
        return self.tr("Eliminate selected polygons")

    def shortDescription(self):
        return self.tr(
            "Combines selected polygons from the input layer with certain adjacent polygons "
            "by erasing their common boundary."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm combines selected polygons from the input layer with certain adjacent polygons "
            "by erasing their common boundary. The adjacent polygon can be either the one with the largest "
            "or smallest area or the one sharing the largest common boundary with the polygon to be eliminated.\n"
            "Eliminate is normally used to get rid of sliver polygons, i.e. tiny polygons that are a result "
            "of polygon intersection processes where boundaries of the inputs are similar but not identical."
        )

    def processAlgorithm(self, parameters, context, feedback):
        inLayer = self.parameterAsVectorLayer(parameters, self.INPUT, context)
        boundary = (
            self.parameterAsEnum(parameters, self.MODE, context) == self.MODE_BOUNDARY
        )
        smallestArea = (
            self.parameterAsEnum(parameters, self.MODE, context)
            == self.MODE_SMALLEST_AREA
        )

        if inLayer.selectedFeatureCount() == 0:
            feedback.reportError(
                self.tr('{0}: (No selection in input layer "{1}")').format(
                    self.displayName(), parameters[self.INPUT]
                )
            )

        featToEliminate = []
        selFeatIds = inLayer.selectedFeatureIds()

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            inLayer.fields(),
            inLayer.wkbType(),
            inLayer.sourceCrs(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        for aFeat in inLayer.getFeatures():
            if feedback.isCanceled():
                break

            if aFeat.id() in selFeatIds:
                # Keep references to the features to eliminate
                featToEliminate.append(aFeat)
            else:
                # write the others to output
                sink.addFeature(aFeat, QgsFeatureSink.Flag.FastInsert)
                feedback.featureAddedToSink(self.OUTPUT)

        sink.finalize()
        feedback.featureSinkFinalized(self.OUTPUT)
        del sink

        # Delete all features to eliminate in processLayer
        processLayer = QgsProcessingUtils.mapLayerFromString(dest_id, context)
        processLayer.startEditing()

        # ANALYZE
        if len(featToEliminate) > 0:  # Prevent zero division
            start = 20.00
            add = 80.00 / len(featToEliminate)
        else:
            start = 100

        feedback.setProgress(start)
        madeProgress = True

        # We go through the list and see if we find any polygons we can
        # merge the selected with. If we have no success with some we
        # merge and then restart the whole story.
        while madeProgress:  # Check if we made any progress
            madeProgress = False
            featNotEliminated = []

            # Iterate over the polygons to eliminate
            for i in range(len(featToEliminate)):
                if feedback.isCanceled():
                    break

                feat = featToEliminate.pop()
                geom2Eliminate = feat.geometry()
                bbox = geom2Eliminate.boundingBox()
                fit = processLayer.getFeatures(
                    QgsFeatureRequest().setFilterRect(bbox).setSubsetOfAttributes([])
                )
                mergeWithFid = None
                mergeWithGeom = None
                max = 0
                min = -1
                selFeat = QgsFeature()

                # use prepared geometries for faster intersection tests
                engine = QgsGeometry.createGeometryEngine(geom2Eliminate.constGet())
                engine.prepareGeometry()

                while fit.nextFeature(selFeat):
                    if feedback.isCanceled():
                        break

                    selGeom = selFeat.geometry()

                    if engine.intersects(selGeom.constGet()):
                        # We have a candidate
                        iGeom = geom2Eliminate.intersection(selGeom)

                        if not iGeom:
                            continue

                        if boundary:
                            selValue = iGeom.length()
                        else:
                            # area. We need a common boundary in
                            # order to merge
                            if 0 < iGeom.length():
                                selValue = selGeom.area()
                            else:
                                selValue = -1

                        if -1 != selValue:
                            useThis = True

                            if smallestArea:
                                if -1 == min:
                                    min = selValue
                                else:
                                    if selValue < min:
                                        min = selValue
                                    else:
                                        useThis = False
                            else:
                                if selValue > max:
                                    max = selValue
                                else:
                                    useThis = False

                            if useThis:
                                mergeWithFid = selFeat.id()
                                mergeWithGeom = QgsGeometry(selGeom)
                # End while fit

                if mergeWithFid is not None:
                    # A successful candidate
                    newGeom = mergeWithGeom.combine(geom2Eliminate)

                    if processLayer.changeGeometry(mergeWithFid, newGeom):
                        madeProgress = True
                    else:
                        raise QgsProcessingException(
                            self.tr(
                                "Could not replace geometry of feature with id {0}"
                            ).format(mergeWithFid)
                        )

                    start = start + add
                    feedback.setProgress(start)
                else:
                    featNotEliminated.append(feat)

            # End for featToEliminate

            featToEliminate = featNotEliminated

        # End while
        if not processLayer.commitChanges():
            raise QgsProcessingException(self.tr("Could not commit changes"))

        for feature in featNotEliminated:
            if feedback.isCanceled():
                break

            processLayer.dataProvider().addFeature(
                feature, QgsFeatureSink.Flag.FastInsert
            )

        return {self.OUTPUT: dest_id}



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt.QtWidgets import QApplication

from ..html_elems import HtmlParagraph, HtmlTable
from ..info_model import DatabaseInfo


class SLDatabaseInfo(DatabaseInfo):
    def __init__(self, db):
        self.db = db

    def connectionDetails(self):
        tbl = [
            (
                QApplication.translate("DBManagerPlugin", "Filename:"),
                self.db.connector.dbname,
            )
        ]
        return HtmlTable(tbl)

    def generalInfo(self):
        info = self.db.connector.getInfo()
        tbl = [(QApplication.translate("DBManagerPlugin", "SQLite version:"), info[0])]
        return HtmlTable(tbl)

    def spatialInfo(self):
        ret = []

        info = self.db.connector.getSpatialInfo()
        if info is None:
            return

        tbl = [
            (QApplication.translate("DBManagerPlugin", "Library:"), info[0]),
            ("GEOS:", info[1]),
            ("Proj:", info[2]),
        ]
        ret.append(HtmlTable(tbl))

        if not self.db.connector.has_geometry_columns:
            ret.append(
                HtmlParagraph(
                    QApplication.translate(
                        "DBManagerPlugin",
                        "<warning> geometry_columns table doesn't exist!\n"
                        "This table is essential for many GIS applications for enumeration of tables.",
                    )
                )
            )

        return ret

    def privilegesDetails(self):
        return None



"""
/***************************************************************************
Name                 : DB Manager plugin for virtual layers
Date                 : December 2015
copyright            : (C) 2015 by Hugo Mercier
email                : hugo dot mercier at oslandia dot com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

# this will disable the dbplugin if the connector raise an ImportError
from qgis.core import (
    QgsApplication,
    QgsProject,
    QgsVectorLayer,
    QgsVirtualLayerDefinition,
)
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QIcon

from ..plugin import Database, DBPlugin, Table, TableField, VectorTable
from .connector import VLayerConnector


def classFactory():
    return VLayerDBPlugin


class VLayerDBPlugin(DBPlugin):
    @classmethod
    def icon(self):
        return QgsApplication.getThemeIcon("/mIconVirtualLayer.svg")

    def connectionIcon(self):
        return QgsApplication.getThemeIcon("/providerQgis.svg")

    @classmethod
    def typeName(self):
        return "vlayers"

    @classmethod
    def typeNameString(self):
        return QCoreApplication.translate("db_manager", "Virtual Layers")

    @classmethod
    def providerName(self):
        return "virtual"

    @classmethod
    def connectionSettingsKey(self):
        return "vlayers"

    @classmethod
    def connections(self):
        return [
            VLayerDBPlugin(QCoreApplication.translate("db_manager", "Project layers"))
        ]

    def databasesFactory(self, connection, uri):
        return FakeDatabase(connection, uri)

    def database(self):
        return self.db

    # def info( self ):

    def connect(self, parent=None):
        self.connectToUri("qgis")
        return True


class FakeDatabase(Database):
    def __init__(self, connection, uri):
        Database.__init__(self, connection, uri)

    def connectorsFactory(self, uri):
        return VLayerConnector(uri)

    def dataTablesFactory(self, row, db, schema=None):
        return LTable(row, db, schema)

    def vectorTablesFactory(self, row, db, schema=None):
        return LVectorTable(row, db, schema)

    def rasterTablesFactory(self, row, db, schema=None):
        return None

    def info(self):
        from .info_model import LDatabaseInfo

        return LDatabaseInfo(self)

    def sqlResultModel(self, sql, parent):
        from .data_model import LSqlResultModel

        return LSqlResultModel(self, sql, parent)

    def sqlResultModelAsync(self, sql, parent):
        from .data_model import LSqlResultModelAsync

        return LSqlResultModelAsync(self, sql, parent)

    def toSqlLayer(
        self,
        sql,
        geomCol,
        uniqueCol,
        layerName="QueryLayer",
        layerType=None,
        avoidSelectById=False,
        _filter="",
    ):
        df = QgsVirtualLayerDefinition()
        df.setQuery(sql)
        if uniqueCol is not None:
            uniqueCol = uniqueCol.strip('"').replace('""', '"')
            df.setUid(uniqueCol)
        if geomCol is not None:
            df.setGeometryField(geomCol)
        vl = QgsVectorLayer(df.toString(), layerName, "virtual")
        if _filter:
            vl.setSubsetString(_filter)
        return vl

    def registerDatabaseActions(self, mainWindow):
        return

    def runAction(self, action):
        return

    def uniqueIdFunction(self):
        return None

    def explicitSpatialIndex(self):
        return True

    def spatialIndexClause(self, src_table, src_column, dest_table, dest_column):
        return f'"{src_table}"._search_frame_ = "{dest_table}"."{dest_column}"'

    def supportsComment(self):
        return False


class LTable(Table):
    def __init__(self, row, db, schema=None):
        Table.__init__(self, db, None)
        self.name, self.isView, self.isSysTable = row

    def tableFieldsFactory(self, row, table):
        return LTableField(row, table)

    def tableDataModel(self, parent):
        from .data_model import LTableDataModel

        return LTableDataModel(self, parent)

    def canBeAddedToCanvas(self):
        return False


class LVectorTable(LTable, VectorTable):
    def __init__(self, row, db, schema=None):
        LTable.__init__(self, row[:-5], db, schema)
        VectorTable.__init__(self, db, schema)
        # SpatiaLite does case-insensitive checks for table names, but the
        # SL provider didn't do the same in QGIS < 1.9, so self.geomTableName
        # stores the table name like stored in the geometry_columns table
        self.geomTableName, self.geomColumn, self.geomType, self.geomDim, self.srid = (
            row[-5:]
        )

    def uri(self):
        uri = self.database().uri()
        uri.setDataSource("", self.geomTableName, self.geomColumn)
        return uri

    def hasSpatialIndex(self, geom_column=None):
        return True

    def createSpatialIndex(self, geom_column=None):
        return

    def deleteSpatialIndex(self, geom_column=None):
        return

    def refreshTableEstimatedExtent(self):
        self.extent = self.database().connector.getTableExtent(
            ("id", self.geomTableName), None
        )

    def runAction(self, action):
        return

    def toMapLayer(self, geometryType=None, crs=None):
        return QgsProject.instance().mapLayer(self.geomTableName)


class LTableField(TableField):
    def __init__(self, row, table):
        TableField.__init__(self, table)
        (
            self.num,
            self.name,
            self.dataType,
            self.notNull,
            self.default,
            self.primaryKey,
        ) = row
        self.hasDefault = self.default



"""
***************************************************************************
    rasterize_over_fixed_value.py
    ---------------------
    Date                 : September 2013
    Copyright            : (C) 2013 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "September 2013"
__copyright__ = "(C) 2013, Alexander Bruy"

import os

from qgis.core import (
    QgsProcessingException,
    QgsProcessingOutputRasterLayer,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class rasterize_over_fixed_value(GdalAlgorithm):
    INPUT = "INPUT"
    INPUT_RASTER = "INPUT_RASTER"
    ADD = "ADD"
    EXTRA = "EXTRA"
    BURN = "BURN"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT, self.tr("Input vector layer")
            )
        )
        self.addParameter(
            QgsProcessingParameterRasterLayer(
                self.INPUT_RASTER, self.tr("Input raster layer")
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.BURN,
                self.tr("A fixed value to burn"),
                type=QgsProcessingParameterNumber.Type.Double,
                defaultValue=0.0,
            )
        )

        params = [
            QgsProcessingParameterBoolean(
                self.ADD,
                self.tr("Add burn in values to existing raster values"),
                defaultValue=False,
            ),
            QgsProcessingParameterString(
                self.EXTRA,
                self.tr("Additional command-line parameters"),
                defaultValue=None,
                optional=True,
            ),
        ]
        for p in params:
            p.setFlags(p.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced)
            self.addParameter(p)

        self.addOutput(
            QgsProcessingOutputRasterLayer(self.OUTPUT, self.tr("Rasterized"))
        )

    def name(self):
        return "rasterize_over_fixed_value"

    def displayName(self):
        return self.tr("Rasterize (overwrite with fixed value)")

    def group(self):
        return self.tr("Vector conversion")

    def groupId(self):
        return "vectorconversion"

    def icon(self):
        return QIcon(os.path.join(pluginPath, "images", "gdaltools", "rasterize.png"))

    def commandName(self):
        return "gdal_rasterize"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        input_details = self.getOgrCompatibleSource(
            self.INPUT, parameters, context, feedback, executing
        )
        inLayer = self.parameterAsRasterLayer(parameters, self.INPUT_RASTER, context)
        if inLayer is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT_RASTER)
            )
        input_raster_details = GdalUtils.gdal_connection_details_from_layer(inLayer)

        self.setOutputValue(self.OUTPUT, inLayer.source())

        arguments = [
            "-l",
            input_details.layer_name,
            "-burn",
            str(self.parameterAsDouble(parameters, self.BURN, context)),
        ]

        if self.parameterAsBool(parameters, self.ADD, context):
            arguments.append("-add")

        if input_details.open_options:
            if GdalUtils.version() < 3070000:
                raise QgsProcessingException(
                    self.tr(
                        "Open options are not supported by gdal_rasterize version {} (requires GDAL version 3.7 or later)"
                    ).format(GdalUtils.readableVersion())
                )

            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        arguments.append(input_details.connection_string)
        arguments.append(input_raster_details.connection_string)

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]

    def postProcessAlgorithm(self, context, feedback):
        fileName = self.output_values.get(self.OUTPUT)
        if not fileName:
            return {}

        if context.project():
            for l in context.project().mapLayers().values():
                if l.source() != fileName:
                    continue

                l.dataProvider().reloadData()
                l.triggerRepaint()

        for l in context.temporaryLayerStore().mapLayers().values():
            if l.source() != fileName:
                continue

            l.dataProvider().reloadData()

        return {}



"""
/***************************************************************************
                           qgsplugininstallerfetchingdialog.py
                           Plugin Installer module
                             -------------------
    Date                 : June 2013
    Copyright            : (C) 2013 by Borys Jurgiel
    Email                : info at borysjurgiel dot pl

    This module is based on former plugin_installer plugin:
      Copyright (C) 2007-2008 Matthew Perry
      Copyright (C) 2008-2013 Borys Jurgiel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from pathlib import Path

from qgis.gui import QgsGui
from qgis.PyQt import uic
from qgis.PyQt.QtCore import QCoreApplication, Qt
from qgis.PyQt.QtWidgets import QDialog, QTreeWidgetItem

from .installer_data import repositories

Ui_QgsPluginInstallerFetchingDialogBase, _ = uic.loadUiType(
    Path(__file__).parent / "qgsplugininstallerfetchingbase.ui"
)


class QgsPluginInstallerFetchingDialog(
    QDialog, Ui_QgsPluginInstallerFetchingDialogBase
):
    # ----------------------------------------- #

    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.setupUi(self)
        QgsGui.instance().enableAutoGeometryRestore(self)

        self.progressBar.setRange(0, len(repositories.allEnabled()) * 100)
        self.itemProgress = {}
        self.item = {}
        for key in repositories.allEnabled():
            self.item[key] = QTreeWidgetItem(self.treeWidget)
            self.item[key].setText(0, key)
            if repositories.all()[key]["state"] > 1:
                self.itemProgress[key] = 100
                self.displayState(key, 0)
            else:
                self.itemProgress[key] = 0
                self.displayState(key, 2)
        self.treeWidget.resizeColumnToContents(0)
        self.treeWidget.setCursor(Qt.CursorShape.WaitCursor)
        repositories.repositoryFetched.connect(self.repositoryFetched)
        repositories.anythingChanged.connect(self.displayState)

    # ----------------------------------------- #
    def displayState(self, key, state, state2=None):
        messages = [
            self.tr("Success"),
            QCoreApplication.translate(
                "QgsPluginInstallerFetchingDialog", "Resolving host name…"
            ),
            QCoreApplication.translate(
                "QgsPluginInstallerFetchingDialog", "Connecting…"
            ),
            QCoreApplication.translate(
                "QgsPluginInstallerFetchingDialog", "Host connected. Sending request…"
            ),
            QCoreApplication.translate(
                "QgsPluginInstallerFetchingDialog", "Downloading data…"
            ),
            self.tr("Idle"),
            QCoreApplication.translate(
                "QgsPluginInstallerFetchingDialog", "Closing connection…"
            ),
            self.tr("Error"),
        ]
        message = messages[state]
        if state2:
            message += " (%s%%)" % state2
        self.item[key].setText(1, message)

        if state == 4 and state2:
            self.itemProgress[key] = state2
        totalProgress = sum(self.itemProgress.values())
        self.progressBar.setValue(totalProgress)

    # ----------------------------------------- #
    def repositoryFetched(self, repoName):
        self.itemProgress[repoName] = 100
        if repositories.all()[repoName]["state"] == 2:
            self.displayState(repoName, 0)
        else:
            self.displayState(repoName, 7)
        if not repositories.fetchingInProgress():
            self.close()



"""
/***************************************************************************
Name                 : TopoViewer plugin for DB Manager
Description          : Create a project to display topology schema on Qgis
Date                 : Sep 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
                       (C) 2019 by Sandro Santilli
email                : strk@kbt.io

Based on qgis_pgis_topoview by Sandro Santilli <strk@kbt.io>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os

from qgis.core import Qgis, QgsLayerTreeGroup, QgsProject, QgsVectorLayer, QgsWkbTypes
from qgis.gui import QgsMessageBar
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction

current_path = os.path.dirname(__file__)


# The load function is called when the "db" database or either one of its
# children db objects (table o schema) is selected by the user.
# @param db is the selected database
# @param mainwindow is the DBManager mainwindow


def load(db, mainwindow):
    # check whether the selected database supports topology
    # (search for topology.topology)
    sql = """SELECT count(*)
                FROM pg_class AS cls JOIN pg_namespace AS nsp ON nsp.oid = cls.relnamespace
                WHERE cls.relname = 'topology' AND nsp.nspname = 'topology'"""
    res = db.executeSql(sql)
    if res is None or len(res) < 1 or int(res[0][0]) <= 0:
        return

    # add the action to the DBManager menu
    action = QAction(QIcon(), "&TopoViewer", db)
    mainwindow.registerAction(action, "&Schema", run)


# The run function is called once the user clicks on the action TopoViewer
# (look above at the load function) from the DBManager menu/toolbar.
# @param item is the selected db item (either db, schema or table)
# @param action is the clicked action on the DBManager menu/toolbar
# @param mainwindow is the DBManager mainwindow
def run(item, action, mainwindow):
    db = item.database()
    uri = db.uri()
    iface = mainwindow.iface

    quoteId = db.connector.quoteId
    quoteStr = db.connector.quoteString

    # check if the selected item is a topology schema
    isTopoSchema = False

    if not hasattr(item, "schema"):
        mainwindow.infoBar.pushMessage(
            "Invalid topology",
            "Select a topology schema to continue.",
            Qgis.MessageLevel.Info,
            mainwindow.iface.messageTimeout(),
        )
        return False

    if item.schema() is not None:
        sql = "SELECT srid FROM topology.topology WHERE name = %s" % quoteStr(
            item.schema().name
        )
        res = db.executeSql(sql)
        isTopoSchema = len(res) > 0

    if not isTopoSchema:
        mainwindow.infoBar.pushMessage(
            "Invalid topology",
            f'Schema "{item.schema().name}" is not registered in topology.topology.',
            Qgis.MessageLevel.Warning,
            mainwindow.iface.messageTimeout(),
        )
        return False

    if res[0][0] < 0:
        mainwindow.infoBar.pushMessage(
            "WARNING",
            f'Topology "{item.schema().name}" is registered as having a srid of {res[0]} in topology.topology, we will assume 0 (for unknown)',
            Qgis.MessageLevel.Warning,
            mainwindow.iface.messageTimeout(),
        )
        toposrid = "0"
    else:
        toposrid = str(res[0][0])

    # load layers into the current project
    toponame = item.schema().name
    template_dir = os.path.join(current_path, "templates")

    # do not refresh the canvas until all the layers are added
    wasFrozen = iface.mapCanvas().isFrozen()
    iface.mapCanvas().freeze()
    try:
        provider = db.dbplugin().providerName()
        uri = db.uri()

        # Force use of estimated metadata (topologies can be big)
        uri.setUseEstimatedMetadata(True)

        # FACES

        # face mbr
        uri.setDataSource(toponame, "face", "mbr", "", "face_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.Polygon)
        layerFaceMbr = QgsVectorLayer(
            uri.uri(False), "%s.face_mbr" % toponame, provider
        )
        layerFaceMbr.loadNamedStyle(os.path.join(template_dir, "face_mbr.qml"))

        face_extent = layerFaceMbr.extent()

        # face geometry
        sql = (
            "SELECT face_id, mbr, topology.ST_GetFaceGeometry(%s,"
            "face_id)::geometry(polygon, %s) as geom "
            "FROM %s.face WHERE face_id > 0"
            % (quoteStr(toponame), toposrid, quoteId(toponame))
        )
        uri.setDataSource("", "(%s\n)" % sql, "geom", "", "face_id")
        uri.setParam("bbox", "mbr")
        uri.setParam("checkPrimaryKeyUnicity", "0")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.Polygon)
        layerFaceGeom = QgsVectorLayer(uri.uri(False), "%s.face" % toponame, provider)
        layerFaceGeom.setExtent(face_extent)
        layerFaceGeom.loadNamedStyle(os.path.join(template_dir, "face.qml"))

        # face_seed
        sql = (
            "SELECT face_id, mbr, ST_PointOnSurface("
            "topology.ST_GetFaceGeometry(%s,"
            "face_id))::geometry(point, %s) as geom "
            "FROM %s.face WHERE face_id > 0"
            % (quoteStr(toponame), toposrid, quoteId(toponame))
        )
        uri.setDataSource("", "(%s)" % sql, "geom", "", "face_id")
        uri.setParam("bbox", "mbr")
        uri.setParam("checkPrimaryKeyUnicity", "0")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.Point)
        layerFaceSeed = QgsVectorLayer(
            uri.uri(False), "%s.face_seed" % toponame, provider
        )
        layerFaceSeed.setExtent(face_extent)
        layerFaceSeed.loadNamedStyle(os.path.join(template_dir, "face_seed.qml"))

        # TODO: add polygon0, polygon1 and polygon2 ?

        # NODES

        # node
        uri.setDataSource(toponame, "node", "geom", "", "node_id")
        uri.removeParam("bbox")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.Point)
        layerNode = QgsVectorLayer(uri.uri(False), "%s.node" % toponame, provider)
        layerNode.loadNamedStyle(os.path.join(template_dir, "node.qml"))
        node_extent = layerNode.extent()

        # node labels
        uri.setDataSource(toponame, "node", "geom", "", "node_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.Point)
        uri.removeParam("bbox")
        layerNodeLabel = QgsVectorLayer(
            uri.uri(False), "%s.node_id" % toponame, provider
        )
        layerNodeLabel.setExtent(node_extent)
        layerNodeLabel.loadNamedStyle(os.path.join(template_dir, "node_label.qml"))

        # EDGES

        # edge
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerEdge = QgsVectorLayer(uri.uri(False), "%s.edge" % toponame, provider)
        edge_extent = layerEdge.extent()

        # directed edge
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerDirectedEdge = QgsVectorLayer(
            uri.uri(False), "%s.directed_edge" % toponame, provider
        )
        layerDirectedEdge.setExtent(edge_extent)
        layerDirectedEdge.loadNamedStyle(os.path.join(template_dir, "edge.qml"))

        # edge labels
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerEdgeLabel = QgsVectorLayer(
            uri.uri(False), "%s.edge_id" % toponame, provider
        )
        layerEdgeLabel.setExtent(edge_extent)
        layerEdgeLabel.loadNamedStyle(os.path.join(template_dir, "edge_label.qml"))

        # face_left
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerFaceLeft = QgsVectorLayer(
            uri.uri(False), "%s.face_left" % toponame, provider
        )
        layerFaceLeft.setExtent(edge_extent)
        layerFaceLeft.loadNamedStyle(os.path.join(template_dir, "face_left.qml"))

        # face_right
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerFaceRight = QgsVectorLayer(
            uri.uri(False), "%s.face_right" % toponame, provider
        )
        layerFaceRight.setExtent(edge_extent)
        layerFaceRight.loadNamedStyle(os.path.join(template_dir, "face_right.qml"))

        # next_left
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerNextLeft = QgsVectorLayer(
            uri.uri(False), "%s.next_left" % toponame, provider
        )
        layerNextLeft.setExtent(edge_extent)
        layerNextLeft.loadNamedStyle(os.path.join(template_dir, "next_left.qml"))

        # next_right
        uri.setDataSource(toponame, "edge_data", "geom", "", "edge_id")
        uri.setSrid(toposrid)
        uri.setWkbType(QgsWkbTypes.Type.LineString)
        uri.removeParam("bbox")
        layerNextRight = QgsVectorLayer(
            uri.uri(False), "%s.next_right" % toponame, provider
        )
        layerNextRight.setExtent(edge_extent)
        layerNextRight.loadNamedStyle(os.path.join(template_dir, "next_right.qml"))

        # Add layers to the layer tree

        faceLayers = [layerFaceMbr, layerFaceGeom, layerFaceSeed]
        nodeLayers = [layerNode, layerNodeLabel]
        edgeLayers = [
            layerEdge,
            layerDirectedEdge,
            layerEdgeLabel,
            layerFaceLeft,
            layerFaceRight,
            layerNextLeft,
            layerNextRight,
        ]

        QgsProject.instance().addMapLayers(faceLayers, False)
        QgsProject.instance().addMapLayers(nodeLayers, False)
        QgsProject.instance().addMapLayers(edgeLayers, False)

        # Organize layers in groups

        groupFaces = QgsLayerTreeGroup("Faces")
        for layer in faceLayers:
            nodeLayer = groupFaces.addLayer(layer)
            nodeLayer.setItemVisibilityChecked(False)
            nodeLayer.setExpanded(False)

        groupNodes = QgsLayerTreeGroup("Nodes")
        for layer in nodeLayers:
            nodeLayer = groupNodes.addLayer(layer)
            nodeLayer.setItemVisibilityChecked(False)
            nodeLayer.setExpanded(False)

        groupEdges = QgsLayerTreeGroup("Edges")
        for layer in edgeLayers:
            nodeLayer = groupEdges.addLayer(layer)
            nodeLayer.setItemVisibilityChecked(False)
            nodeLayer.setExpanded(False)

        supergroup = QgsLayerTreeGroup('Topology "%s"' % toponame)
        supergroup.insertChildNodes(-1, [groupFaces, groupNodes, groupEdges])

        layerTree = QgsProject.instance().layerTreeRoot()

        layerTree.addChildNode(supergroup)

        # Set layers rendering order

        order = layerTree.layerOrder()

        order.insert(0, order.pop(order.index(layerFaceMbr)))
        order.insert(0, order.pop(order.index(layerFaceGeom)))
        order.insert(0, order.pop(order.index(layerEdge)))
        order.insert(0, order.pop(order.index(layerDirectedEdge)))

        order.insert(0, order.pop(order.index(layerNode)))
        order.insert(0, order.pop(order.index(layerFaceSeed)))

        order.insert(0, order.pop(order.index(layerNodeLabel)))
        order.insert(0, order.pop(order.index(layerEdgeLabel)))

        order.insert(0, order.pop(order.index(layerNextLeft)))
        order.insert(0, order.pop(order.index(layerNextRight)))
        order.insert(0, order.pop(order.index(layerFaceLeft)))
        order.insert(0, order.pop(order.index(layerFaceRight)))

        layerTree.setHasCustomLayerOrder(True)
        layerTree.setCustomLayerOrder(order)

    finally:
        # Set canvas extent to topology extent, if not yet initialized
        canvas = iface.mapCanvas()
        if canvas.fullExtent().isNull():
            ext = node_extent
            ext.combineExtentWith(edge_extent)
            # Grow by 1/20 of largest side
            ext = ext.buffered(max(ext.width(), ext.height()) / 20)
            canvas.setExtent(ext)

        # restore canvas render flag
        if not wasFrozen:
            iface.mapCanvas().freeze(False)

    return True



"""
/***************************************************************************
    KNearestConcaveHull.py
    ----------------------
    Date                 : November 2014
    Copyright            : (C) 2014 by Detlev Neumann
                           Dr. Neumann Consulting - Geospatial Services
    Email                : dneumann@geospatial-services.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = "Detlev Neumann"
__date__ = "November 2014"
__copyright__ = "(C) 2014, Detlev Neumann"

import math
import os.path

from qgis.core import (
    QgsApplication,
    QgsExpression,
    QgsFeature,
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsPoint,
    QgsPointXY,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterNumber,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QMetaType
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class KNearestConcaveHull(QgisAlgorithm):
    KNEIGHBORS = "KNEIGHBORS"
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    FIELD = "FIELD"

    def name(self):
        return "knearestconcavehull"

    def displayName(self):
        return self.tr("Concave hull (k-nearest neighbor)")

    def shortDescription(self):
        return self.tr("Generates a concave hull polygon from a set of points.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm generates a concave hull polygon from a set of points. "
            "If the input layer is a line or polygon layer, it will use the nodes.\n"
            "The number of neighbours to consider determines the concaveness of the output polygon. "
            "A lower number will result in a concave hull that follows the points very closely, "
            "while a higher number will have a smoother shape. The minimum number of neighbour points to consider is 3. "
            "A value equal to or greater than the number of points will result in a convex hull.\n"
            "If a field is selected, the algorithm will group the features in the input layer "
            "using unique values in that field and generate individual polygons in the output layer for each group."
        )

    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmConcaveHull.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmConcaveHull.svg")

    def group(self):
        return self.tr("Vector geometry")

    def groupId(self):
        return "vectorgeometry"

    def flags(self):
        return (
            super().flags()
            | QgsProcessingAlgorithm.Flag.FlagDeprecated
            | QgsProcessingAlgorithm.Flag.FlagNotAvailableInStandaloneTool
        )

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.KNEIGHBORS,
                self.tr(
                    "Number of neighboring points to consider (a lower number is more concave, a higher number is smoother)"
                ),
                QgsProcessingParameterNumber.Type.Integer,
                defaultValue=3,
                minValue=3,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.FIELD,
                self.tr("Field (set if creating concave hulls by class)"),
                parentLayerParameterName=self.INPUT,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Concave hull"),
                QgsProcessing.SourceType.TypeVectorPolygon,
            )
        )

    def processAlgorithm(self, parameters, context, feedback):
        # Get variables from dialog
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        field_name = self.parameterAsString(parameters, self.FIELD, context)
        kneighbors = self.parameterAsInt(parameters, self.KNEIGHBORS, context)
        use_field = bool(field_name)

        field_index = -1

        fields = QgsFields()
        fields.append(QgsField("id", QMetaType.Type.Int, "", 20))

        current = 0

        # Get properties of the field the grouping is based on
        if use_field:
            field_index = source.fields().lookupField(field_name)
            if field_index >= 0:
                fields.append(
                    source.fields()[field_index]
                )  # Add a field with the name of the grouping field

                # Initialize writer
                (sink, dest_id) = self.parameterAsSink(
                    parameters,
                    self.OUTPUT,
                    context,
                    fields,
                    QgsWkbTypes.Type.Polygon,
                    source.sourceCrs(),
                )
                if sink is None:
                    raise QgsProcessingException(
                        self.invalidSinkError(parameters, self.OUTPUT)
                    )

                success = False
                fid = 0

                # Get unique values of grouping field
                unique_values = source.uniqueValues(field_index)
                total = 100.0 / float(source.featureCount() * len(unique_values))

                for unique in unique_values:
                    points = []
                    filter = QgsExpression.createFieldEqualityExpression(
                        field_name, unique
                    )
                    request = QgsFeatureRequest().setFilterExpression(filter)
                    request.setSubsetOfAttributes([])
                    # Get features with the grouping attribute equal to the current grouping value
                    features = source.getFeatures(request)
                    for in_feature in features:
                        if feedback.isCanceled():
                            break
                        # Add points or vertices of more complex geometry
                        points.extend(extract_points(in_feature.geometry()))
                        current += 1
                        feedback.setProgress(int(current * total))
                    # A minimum of 3 points is necessary to proceed
                    if len(points) >= 3:
                        out_feature = QgsFeature()
                        the_hull = concave_hull(points, kneighbors)
                        if the_hull:
                            vertex = [
                                QgsPointXY(point[0], point[1]) for point in the_hull
                            ]
                            poly = QgsGeometry().fromPolygonXY([vertex])

                            out_feature.setGeometry(poly)
                            # Give the polygon the same attribute as the point grouping attribute
                            out_feature.setAttributes([fid, unique])
                            sink.addFeature(out_feature, QgsFeatureSink.Flag.FastInsert)
                            feedback.featureAddedToSink(self.OUTPUT)
                            success = True  # at least one polygon created
                    fid += 1
                if not success:
                    raise QgsProcessingException(
                        "No hulls could be created. Most likely there were not at least three unique points in any of the groups."
                    )
                sink.finalize()
                feedback.featureSinkFinalized(self.OUTPUT)
            else:
                # Field parameter provided but can't read from it
                raise QgsProcessingException("Unable to find grouping field")

        else:
            # Not grouped by field
            # Initialize writer
            (sink, dest_id) = self.parameterAsSink(
                parameters,
                self.OUTPUT,
                context,
                fields,
                QgsWkbTypes.Type.Polygon,
                source.sourceCrs(),
            )
            if sink is None:
                raise QgsProcessingException(
                    self.invalidSinkError(parameters, self.OUTPUT)
                )

            points = []
            request = QgsFeatureRequest()
            request.setSubsetOfAttributes([])
            features = source.getFeatures(request)  # Get all features
            total = 100.0 / source.featureCount() if source.featureCount() else 0
            for in_feature in features:
                if feedback.isCanceled():
                    break
                # Add points or vertices of more complex geometry
                points.extend(extract_points(in_feature.geometry()))
                current += 1
                feedback.setProgress(int(current * total))

            # A minimum of 3 points is necessary to proceed
            if len(points) >= 3:
                out_feature = QgsFeature()
                the_hull = concave_hull(points, kneighbors)
                if the_hull:
                    vertex = [QgsPointXY(point[0], point[1]) for point in the_hull]
                    poly = QgsGeometry().fromPolygonXY([vertex])

                    out_feature.setGeometry(poly)
                    out_feature.setAttributes([0])
                    sink.addFeature(out_feature, QgsFeatureSink.Flag.FastInsert)
                    feedback.featureAddedToSink(self.OUTPUT)
                else:
                    # the_hull returns None only when there are less than three points after cleaning
                    raise QgsProcessingException(
                        "At least three unique points are required to create a concave hull."
                    )
            else:
                raise QgsProcessingException(
                    "At least three points are required to create a concave hull."
                )

            sink.finalize()
            feedback.featureSinkFinalized(self.OUTPUT)

        return {self.OUTPUT: dest_id}


def clean_list(list_of_points):
    """
    Deletes duplicate points in list_of_points
    """
    return list(set(list_of_points))


def find_min_y_point(list_of_points):
    """
    Returns that point of *list_of_points* having minimal y-coordinate

    :param list_of_points: list of tuples
    :return: tuple (x, y)
    """
    min_y_pt = list_of_points[0]
    for point in list_of_points[1:]:
        if point[1] < min_y_pt[1] or (
            point[1] == min_y_pt[1] and point[0] < min_y_pt[0]
        ):
            min_y_pt = point
    return min_y_pt


def add_point(vector, element):
    """
    Returns vector with the given element append to the right
    """
    vector.append(element)
    return vector


def remove_point(vector, element):
    """
    Returns a copy of vector without the given element
    """
    vector.pop(vector.index(element))
    return vector


def euclidean_distance(point1, point2):
    """
    Returns the euclidean distance of the 2 given points.

    :param point1: tuple (x, y)
    :param point2: tuple (x, y)
    :return: float
    """
    return math.sqrt(
        math.pow(point1[0] - point2[0], 2) + math.pow(point1[1] - point2[1], 2)
    )


def nearest_points(list_of_points, point, k):
    """
    Returns a list of the indices of the k closest neighbors from list_of_points to the specified point. The measure
    of proximity is the Euclidean distance. Internally, k becomes the minimum between the given value for k and the
    number of points in list_of_points

    :param list_of_points: list of tuples
    :param point: tuple (x, y)
    :param k: integer
    :return: list of k tuples
    """
    # build a list of tuples of distances between point *point* and every point in *list_of_points*, and
    # their respective index of list *list_of_distances*
    list_of_distances = []
    for index in range(len(list_of_points)):
        list_of_distances.append(
            (euclidean_distance(list_of_points[index], point), index)
        )

    # sort distances in ascending order
    list_of_distances.sort()

    # get the k nearest neighbors of point
    nearest_list = []
    for index in range(min(k, len(list_of_points))):
        nearest_list.append(list_of_points[list_of_distances[index][1]])
    return nearest_list


def angle(from_point, to_point):
    """
    Returns the angle of the directed line segment, going from *from_point* to *to_point*, in radians. The angle is
    positive for segments with upward direction (north), otherwise negative (south). Values ranges from 0 at the
    right (east) to pi at the left side (west).

    :param from_point: tuple (x, y)
    :param to_point: tuple (x, y)
    :return: float
    """
    return math.atan2(to_point[1] - from_point[1], to_point[0] - from_point[0])


def angle_difference(angle1, angle2):
    """
    Calculates the difference between the given angles in clockwise direction as radians.

    :param angle1: float
    :param angle2: float
    :return: float; between 0 and 2*Pi
    """
    if (angle1 > 0 and angle2 >= 0) and angle1 > angle2:
        return abs(angle1 - angle2)
    elif (angle1 >= 0 and angle2 > 0) and angle1 < angle2:
        return 2 * math.pi + angle1 - angle2
    elif (angle1 < 0 and angle2 <= 0) and angle1 < angle2:
        return 2 * math.pi + angle1 + abs(angle2)
    elif (angle1 <= 0 and angle2 < 0) and angle1 > angle2:
        return abs(angle1 - angle2)
    elif angle1 <= 0 < angle2:
        return 2 * math.pi + angle1 - angle2
    elif angle1 >= 0 >= angle2:
        return angle1 + abs(angle2)
    else:
        return 0


def intersect(line1, line2):
    """
    Returns True if the two given line segments intersect each other, and False otherwise.

    :param line1: 2-tuple of tuple (x, y)
    :param line2: 2-tuple of tuple (x, y)
    :return: boolean
    """
    a1 = line1[1][1] - line1[0][1]
    b1 = line1[0][0] - line1[1][0]
    c1 = a1 * line1[0][0] + b1 * line1[0][1]
    a2 = line2[1][1] - line2[0][1]
    b2 = line2[0][0] - line2[1][0]
    c2 = a2 * line2[0][0] + b2 * line2[0][1]
    tmp = a1 * b2 - a2 * b1
    if tmp == 0:
        return False
    sx = (c1 * b2 - c2 * b1) / tmp
    if (
        (sx > line1[0][0] and sx > line1[1][0])
        or (sx > line2[0][0] and sx > line2[1][0])
        or (sx < line1[0][0] and sx < line1[1][0])
        or (sx < line2[0][0] and sx < line2[1][0])
    ):
        return False
    sy = (a1 * c2 - a2 * c1) / tmp
    if (
        (sy > line1[0][1] and sy > line1[1][1])
        or (sy > line2[0][1] and sy > line2[1][1])
        or (sy < line1[0][1] and sy < line1[1][1])
        or (sy < line2[0][1] and sy < line2[1][1])
    ):
        return False
    return True


def point_in_polygon_q(point, list_of_points):
    """
    Return True if given point *point* is laying in the polygon described by the vertices *list_of_points*,
    otherwise False

    Based on the "Ray Casting Method" described by Joel Lawhead in this blog article:
    http://geospatialpython.com/2011/01/point-in-polygon.html

    """
    x = point[0]
    y = point[1]
    poly = [(pt[0], pt[1]) for pt in list_of_points]
    n = len(poly)
    inside = False

    p1x, p1y = poly[0]
    for i in range(n + 1):
        p2x, p2y = poly[i % n]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y

    return inside


def extract_points(geom):
    """
    Generate list of QgsPoints from QgsGeometry *geom* ( can be point, line, or polygon )
    Code taken from fTools plugin

    :param geom: an arbitrary geometry feature
    :return: list of points
    """
    multi_geom = QgsGeometry()
    temp_geom = []
    # point geometry
    if geom.type() == 0:
        if geom.isMultipart():
            temp_geom = geom.asMultiPoint()
        else:
            temp_geom.append(geom.asPoint())
    # line geometry
    if geom.type() == 1:
        # if multipart feature explode to single part
        if geom.isMultipart():
            multi_geom = geom.asMultiPolyline()
            for i in multi_geom:
                temp_geom.extend(i)
        else:
            temp_geom = geom.asPolyline()
    # polygon geometry
    elif geom.type() == 2:
        # if multipart feature explode to single part
        if geom.isMultipart():
            multi_geom = geom.asMultiPolygon()
            # now single part polygons
            for i in multi_geom:
                # explode to line segments
                for j in i:
                    temp_geom.extend(j)
        else:
            multi_geom = geom.asPolygon()
            # explode to line segments
            for i in multi_geom:
                temp_geom.extend(i)
    return temp_geom


def sort_by_angle(list_of_points, last_point, last_angle):
    """
    returns the points in list_of_points in descending order of angle to the last segment of the envelope, measured
    in a clockwise direction. Thus, the rightmost of the neighboring points is always selected. The first point of
    this list will be the next point of the envelope.
    """

    def getkey(item):
        return angle_difference(last_angle, angle(last_point, item))

    vertex_list = sorted(list_of_points, key=getkey, reverse=True)
    return vertex_list


def concave_hull(points_list, k):
    """
    Calculates a valid concave hull polygon containing all given points. The algorithm searches for that
    point in the neighborhood of k nearest neighbors which maximizes the rotation angle in clockwise direction
    without intersecting any previous line segments.

    This is an implementation of the algorithm described by Adriano Moreira and Maribel Yasmina Santos:
    CONCAVE HULL: A neighborhood_k-NEAREST NEIGHBORS APPROACH FOR THE COMPUTATION OF THE REGION OCCUPIED BY A SET OF POINTS.
    GRAPP 2007 - International Conference on Computer Graphics Theory and Applications; pp 61-68.

    :param points_list: list of tuples (x, y)
    :param k: integer
    :return: list of tuples (x, y)
    """
    # return an empty list if not enough points are given
    if k > len(points_list):
        k = len(points_list)

    # the number of nearest neighbors k must be greater than or equal to 3
    kk = max(k, 3)

    # delete duplicate points
    point_set = clean_list(points_list)

    # if point_set has less then 3 points no polygon can be created and an empty list will be returned
    if len(point_set) < 3:
        return None

    # if point_set has 3 points then these are already vertices of the hull. Append the first point to
    # close the hull polygon
    if len(point_set) == 3:
        return add_point(point_set, point_set[0])

    # make sure that k neighbors can be found
    kk = min(kk, len(point_set))

    # start with the point having the smallest y-coordinate (most southern point)
    first_point = find_min_y_point(point_set)

    # add this points as the first vertex of the hull
    hull = [first_point]

    # make the first vertex of the hull to the current point
    current_point = first_point

    # remove the point from the point_set, to prevent him being among the nearest points
    point_set = remove_point(point_set, first_point)
    previous_angle = math.pi

    # step counts the number of segments
    step = 2

    # as long as point_set is not empty or search is returning to the starting point
    while (current_point != first_point) or (step == 2) and (len(point_set) > 0):
        # after 3 iterations add the first point to point_set again, otherwise a hull cannot be closed
        if step == 5:
            point_set = add_point(point_set, first_point)

        # search the k nearest neighbors of the current point
        k_nearest_points = nearest_points(point_set, current_point, kk)

        # sort the candidates (neighbors) in descending order of right-hand turn. This way the algorithm progresses
        # in clockwise direction through as many points as possible
        c_points = sort_by_angle(k_nearest_points, current_point, previous_angle)

        its = True
        i = -1

        # search for the nearest point to which the connecting line does not intersect any existing segment
        while its is True and (i < len(c_points) - 1):
            i += 1
            if c_points[i] == first_point:
                last_point = 1
            else:
                last_point = 0
            j = 2
            its = False

            while its is False and (j < len(hull) - last_point):
                its = intersect(
                    (hull[step - 2], c_points[i]),
                    (hull[step - 2 - j], hull[step - 1 - j]),
                )
                j += 1

        # there is no candidate to which the connecting line does not intersect any existing segment, so the
        # for the next candidate fails. The algorithm starts again with an increased number of neighbors
        if its is True:
            return concave_hull(points_list, kk + 1)

        # the first point which complies with the requirements is added to the hull and gets the current point
        current_point = c_points[i]
        hull = add_point(hull, current_point)

        # calculate the angle between the last vertex and his precursor, that is the last segment of the hull
        # in reversed direction
        previous_angle = angle(hull[step - 1], hull[step - 2])

        # remove current_point from point_set
        point_set = remove_point(point_set, current_point)

        # increment counter
        step += 1

    all_inside = True
    i = len(point_set) - 1

    # check if all points are within the created polygon
    while (all_inside is True) and (i >= 0):
        all_inside = point_in_polygon_q(point_set[i], hull)
        i -= 1

    # since at least one point is out of the computed polygon, try again with a higher number of neighbors
    if all_inside is False:
        return concave_hull(points_list, kk + 1)

    # a valid hull has been constructed
    return hull



"""
***************************************************************************
    TopoColors.py
    --------------
    Date                 : February 2017
    Copyright            : (C) 2017 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "February 2017"
__copyright__ = "(C) 2017, Nyall Dawson"

import operator
import os
import sys
from collections import defaultdict

from qgis.core import (
    NULL,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsPointXY,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterDistance,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterNumber,
    QgsProcessingUtils,
    QgsSpatialIndex,
)
from qgis.PyQt.QtCore import QMetaType

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class TopoColor(QgisAlgorithm):
    INPUT = "INPUT"
    MIN_COLORS = "MIN_COLORS"
    MIN_DISTANCE = "MIN_DISTANCE"
    BALANCE = "BALANCE"
    OUTPUT = "OUTPUT"

    def tags(self):
        return self.tr("topocolor,colors,graph,adjacent,assign").split(",")

    def group(self):
        return self.tr("Cartography")

    def groupId(self):
        return "cartography"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):

        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input layer"),
                [QgsProcessing.SourceType.TypeVectorPolygon],
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.MIN_COLORS,
                self.tr("Minimum number of colors"),
                minValue=1,
                maxValue=1000,
                defaultValue=4,
            )
        )
        self.addParameter(
            QgsProcessingParameterDistance(
                self.MIN_DISTANCE,
                self.tr("Minimum distance between features"),
                parentParameterName=self.INPUT,
                minValue=0.0,
                defaultValue=0.0,
            )
        )
        balance_by = [
            self.tr("By feature count"),
            self.tr("By assigned area"),
            self.tr("By distance between colors"),
        ]
        self.addParameter(
            QgsProcessingParameterEnum(
                self.BALANCE,
                self.tr("Balance color assignment"),
                options=balance_by,
                defaultValue=0,
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Colored"),
                QgsProcessing.SourceType.TypeVectorPolygon,
            )
        )

    def name(self):
        return "topologicalcoloring"

    def displayName(self):
        return self.tr("Topological coloring")

    def shortDescription(self):
        return self.tr(
            "Assigns a color index to polygon features, so no adjacent polygons share the same color index."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm assigns a color index to polygon features in such a way that "
            "no adjacent polygons share the same color index, whilst minimizing the number of colors required.\n"
            "An optional minimum distance between features assigned the same color can be set to prevent nearby "
            "(but non-touching) features from being assigned equal colors.\n"
            "The algorithm allows choice of method to use when assigning colors. "
            "The default method attempts to assign colors so that the count of features assigned "
            "to each individual color index is balanced.\n"
            "The 'by assigned area' mode instead assigns colors so that the total area of features "
            "assigned to each color is balanced. This mode can be useful to help avoid large features "
            "resulting in one of the colors appearing more dominant on a colored map.\n"
            "The 'by distance between colors' mode will assign colors in order to maximize the distance "
            "between features of the same color. This mode helps to create a more uniform distribution of colors across a map.\n"
            "A minimum number of colors can be specified if desired. The color index is saved to a new attribute named color_id."
        )

    def processAlgorithm(self, parameters, context, feedback):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        min_colors = self.parameterAsInt(parameters, self.MIN_COLORS, context)
        balance_by = self.parameterAsEnum(parameters, self.BALANCE, context)
        min_distance = self.parameterAsDouble(parameters, self.MIN_DISTANCE, context)

        newFields = QgsFields()
        newFields.append(QgsField("color_id", QMetaType.Type.Int))
        fields = QgsProcessingUtils.combineFields(source.fields(), newFields)

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            fields,
            source.wkbType(),
            source.sourceCrs(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        features = {f.id(): f for f in source.getFeatures()}

        topology, id_graph = self.compute_graph(
            features, feedback, min_distance=min_distance
        )
        feature_colors = ColoringAlgorithm.balanced(
            features,
            balance=balance_by,
            graph=topology,
            feedback=feedback,
            min_colors=min_colors,
        )

        if len(feature_colors) == 0:
            return {self.OUTPUT: dest_id}

        max_colors = max(feature_colors.values())
        feedback.pushInfo(self.tr("{} colors required").format(max_colors))

        total = 20.0 / len(features)
        current = 0
        for feature_id, input_feature in features.items():
            if feedback.isCanceled():
                break

            output_feature = input_feature
            attributes = input_feature.attributes()
            if feature_id in feature_colors:
                attributes.append(feature_colors[feature_id])
            else:
                attributes.append(NULL)
            output_feature.setAttributes(attributes)

            sink.addFeature(output_feature, QgsFeatureSink.Flag.FastInsert)
            feedback.featureAddedToSink(self.OUTPUT)
            current += 1
            feedback.setProgress(80 + int(current * total))

        sink.finalize()
        feedback.featureSinkFinalized(self.OUTPUT)
        return {self.OUTPUT: dest_id}

    @staticmethod
    def compute_graph(features, feedback, create_id_graph=False, min_distance=0):
        """compute topology from a layer/field"""
        s = Graph(sort_graph=False)
        id_graph = None
        if create_id_graph:
            id_graph = Graph(sort_graph=True)

        # skip features without geometry
        features_with_geometry = {
            f_id: f for (f_id, f) in features.items() if f.hasGeometry()
        }

        total = 70.0 / len(features_with_geometry) if features_with_geometry else 1
        index = QgsSpatialIndex()

        i = 0
        for feature_id, f in features_with_geometry.items():
            if feedback.isCanceled():
                break

            g = f.geometry()
            if min_distance > 0:
                g = g.buffer(min_distance, 5)

            engine = QgsGeometry.createGeometryEngine(g.constGet())
            engine.prepareGeometry()

            feature_bounds = g.boundingBox()
            # grow bounds a little so we get touching features
            feature_bounds.grow(feature_bounds.width() * 0.01)
            intersections = index.intersects(feature_bounds)
            for l2 in intersections:
                f2 = features_with_geometry[l2]
                if engine.intersects(f2.geometry().constGet()):
                    s.add_edge(f.id(), f2.id())
                    s.add_edge(f2.id(), f.id())
                    if id_graph:
                        id_graph.add_edge(f.id(), f2.id())

            index.addFeature(f)
            i += 1
            feedback.setProgress(int(i * total))

        for feature_id, f in features_with_geometry.items():
            if feedback.isCanceled():
                break

            if feature_id not in s.node_edge:
                s.add_edge(feature_id, None)

        return s, id_graph


class ColoringAlgorithm:
    @staticmethod
    def balanced(features, graph, feedback, balance=0, min_colors=4):
        feature_colors = {}
        # start with minimum number of colors in pool
        color_pool = set(range(1, min_colors + 1))

        # calculate count of neighbours
        neighbour_count = defaultdict(int)
        for feature_id, neighbours in graph.node_edge.items():
            neighbour_count[feature_id] += len(neighbours)

        # sort features by neighbour count - we want to handle those with more neighbours first
        sorted_by_count = [
            feature_id
            for feature_id in sorted(
                neighbour_count.items(), key=operator.itemgetter(1), reverse=True
            )
        ]
        # counts for each color already assigned
        color_counts = defaultdict(int)
        color_areas = defaultdict(float)
        for c in color_pool:
            color_counts[c] = 0
            color_areas[c] = 0

        total = 10.0 / len(sorted_by_count) if sorted_by_count else 1
        i = 0

        for feature_id, n in sorted_by_count:
            if feedback.isCanceled():
                break

            # first work out which already assigned colors are adjacent to this feature
            adjacent_colors = {
                feature_colors[neighbour]
                for neighbour in graph.node_edge[feature_id]
                if neighbour in feature_colors
            }

            # from the existing colors, work out which are available (ie non-adjacent)
            available_colors = color_pool.difference(adjacent_colors)

            feature_color = -1
            if len(available_colors) == 0:
                # no existing colors available for this feature, so add new color to pool and repeat
                min_colors += 1
                return ColoringAlgorithm.balanced(
                    features, graph, feedback, balance, min_colors
                )
            else:
                if balance == 0:
                    # choose least used available color
                    counts = [
                        (c, v) for c, v in color_counts.items() if c in available_colors
                    ]
                    feature_color = sorted(counts, key=operator.itemgetter(1))[0][0]
                    color_counts[feature_color] += 1
                elif balance == 1:
                    areas = [
                        (c, v) for c, v in color_areas.items() if c in available_colors
                    ]
                    feature_color = sorted(areas, key=operator.itemgetter(1))[0][0]
                    color_areas[feature_color] += features[feature_id].geometry().area()
                elif balance == 2:
                    min_distances = {c: sys.float_info.max for c in available_colors}
                    this_feature_centroid = (
                        features[feature_id].geometry().centroid().constGet()
                    )

                    # find features for all available colors
                    other_features = {
                        f_id: c
                        for (f_id, c) in feature_colors.items()
                        if c in available_colors
                    }

                    # loop through these, and calculate the minimum distance from this feature to the nearest
                    # feature with each assigned color
                    for other_feature_id, c in other_features.items():
                        if feedback.isCanceled():
                            break

                        other_geometry = features[other_feature_id].geometry()
                        other_centroid = other_geometry.centroid().constGet()

                        distance = this_feature_centroid.distanceSquared(other_centroid)
                        if distance < min_distances[c]:
                            min_distances[c] = distance

                    # choose color such that minimum distance is maximised! ie we want MAXIMAL separation between
                    # features with the same color
                    feature_color = sorted(
                        min_distances, key=min_distances.__getitem__, reverse=True
                    )[0]

            feature_colors[feature_id] = feature_color

            i += 1
            feedback.setProgress(70 + int(i * total))

        return feature_colors


class Graph:
    def __init__(self, sort_graph=True):
        self.sort_graph = sort_graph
        self.node_edge = {}

    def add_edge(self, i, j):
        ij = [i, j]
        if self.sort_graph:
            ij.sort()
        (i, j) = ij
        if i in self.node_edge:
            self.node_edge[i].add(j)
        else:
            self.node_edge[i] = {j}

    def make_full(self):
        g = Graph(sort_graph=False)
        for k in self.node_edge.keys():
            for v in self.node_edge[k]:
                g.add_edge(v, k)
                g.add_edge(k, v)
        return g



"""
***************************************************************************
    r_series_interp.py
    ------------------
    Date                 : February 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "February 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

import os

from qgis.PyQt.QtCore import QCoreApplication

from grassprovider.grass_utils import GrassUtils


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    datapos = alg.parameterAsDouble(parameters, "datapos", context)
    infile = alg.parameterAsString(parameters, "infile", context)
    output = alg.parameterAsString(parameters, "output", context)
    outfile = alg.parameterAsString(parameters, "outfile", context)

    if datapos and infile:
        return False, QCoreApplication.translate(
            "GrassUtils",
            "You need to set either inline data positions or an input data positions file!",
        )
    if output and outfile:
        return False, QCoreApplication.translate(
            "GrassUtils",
            "You need to set either sampling data positions or an output sampling data positions file!",
        )
    if not (datapos or infile or output or outfile):
        return False, QCoreApplication.translate(
            "GrassUtils", "You need to set input and output data positions parameters!"
        )
    return True, None


def processCommand(alg, parameters, context, feedback):
    # We temporary remove the output directory
    alg.processCommand(parameters, context, feedback, True)


def processOutputs(alg, parameters, context, feedback):
    # We take all the outputs and we export them to the output directory
    outputDir = alg.parameterAsString(parameters, "output_dir", context)
    output = alg.parameterAsString(parameters, "output", context)
    outfile = alg.parameterAsString(parameters, "outfile", context)
    outs = []
    if output:
        outs = output.split(",")
    elif outfile:
        # Handle file manually to find the name of the layers
        with open(outfile) as f:
            for line in f:
                if "|" in line:
                    outs.append(line.split("|")[0])

    createOpt = alg.parameterAsString(parameters, alg.GRASS_RASTER_FORMAT_OPT, context)
    metaOpt = alg.parameterAsString(parameters, alg.GRASS_RASTER_FORMAT_META, context)

    for out in outs:
        # We need to export the raster with all its bands and its color table
        fileName = os.path.join(outputDir, out)
        outFormat = GrassUtils.getRasterFormatFromFilename(fileName)
        alg.exportRasterLayer(out, fileName, True, outFormat, createOpt, metaOpt)
